"""Thread-safe enabled state manager for streaming operations.

This module provides a thread-safe manager for controlling enabled/disabled
state with event emission capabilities. Used to coordinate streaming state
across multiple components and threads.
"""

import threading
from asyncio import AbstractEventLoop

from pyee.asyncio import AsyncIOEventEmitter

from neuracore.core.streaming.event_loop_utils import get_running_loop


class EnabledManager(AsyncIOEventEmitter):
    """Thread-safe manager for enabled/disabled state with event notifications.

    Provides a thread-safe way to manage boolean state with automatic event
    emission when the state changes. Extends AsyncIOEventEmitter to support
    event-driven architecture across async and sync contexts.
    """

    DISABLED = "DISABLED"

    def __init__(
        self, initial_state: bool, loop: AbstractEventLoop | None = None
    ) -> None:
        """Initialize the enabled manager.

        Args:
            initial_state: Initial enabled/disabled state
            loop: Optional event loop for async event emission
        """
        self.loop = loop or get_running_loop()
        super().__init__(loop=self.loop)
        self._is_enabled = initial_state
        self.lock = threading.Lock()

    def is_enabled(self) -> bool:
        """Check if the manager is in enabled state.

        Returns:
            bool: True if enabled, False if disabled
        """
        with self.lock:
            return self._is_enabled and not self.loop.is_closed()

    def is_disabled(self) -> bool:
        """Check if the manager is in disabled state.

        Returns:
            bool: True if disabled, False if enabled
        """
        return not self.is_enabled()

    def disable(self) -> None:
        """Disable the manager and emit notification.

        Thread-safely disables the manager, emits a DISABLED event,
        and removes all event listeners to prevent memory leaks.
        If already disabled, this is a no-op.
        """
        with self.lock:
            if not self._is_enabled:
                return
            self._is_enabled = False
            self.emit(self.DISABLED)
            self.remove_all_listeners()

    @staticmethod
    def all_enabled(
        *managers: "EnabledManager", loop: AbstractEventLoop | None = None
    ) -> "EnabledManager":
        """Constructs a new enabled manager.

        the new manager is only enabled if and only if all the provided managers
        are enabled.

        Args:
            *managers: The managers to check.

        Returns:
            A new enabled manager that is only enabled if all the provided
                managers are enabled.
        """
        enabled_count = sum(1 for manager in managers if manager.is_enabled())
        if enabled_count < len(managers):
            return EnabledManager(initial_state=False, loop=loop)

        new_manager = EnabledManager(initial_state=True, loop=loop)

        for manager in managers:
            manager.add_listener(EnabledManager.DISABLED, lambda: new_manager.disable())

        return new_manager

    @staticmethod
    def any_enabled(
        *managers: "EnabledManager", loop: AbstractEventLoop | None = None
    ) -> "EnabledManager":
        """Constructs a new enabled manager.

        The new manager is only enabled if and only if any of the provided
        managers are enabled.

        Args:
            *managers: The managers to check.

        Returns:
            A new enabled manager that is only enabled if any of the provided
                managers are enabled.
        """
        enabled_count = sum(1 for manager in managers if manager.is_enabled())
        if enabled_count == 0:
            return EnabledManager(initial_state=False, loop=loop)

        new_manager = EnabledManager(initial_state=True, loop=loop)

        def on_disable() -> None:
            nonlocal enabled_count
            enabled_count += 1
            if enabled_count <= 0:
                new_manager.disable()

        for manager in managers:
            manager.add_listener(EnabledManager.DISABLED, on_disable)

        return new_manager

    @classmethod
    def derived_manger(
        cls, manager: "EnabledManager", loop: AbstractEventLoop | None = None
    ) -> "EnabledManager":
        """Constructs a new enabled manager that is derived from another.

        Args:
            manager: The manager to derive from.

        Returns:
            A new enabled manager that is enabled if the provided manager is
                enabled.
        """
        return cls.all_enabled(manager, loop=loop)



"""Robot data logging utilities.

This module provides functions for logging various types of robot sensor data
including joint positions, camera images, point clouds, and custom data streams.
All logging functions support optional robot identification and timestamping.
"""

import time
from warnings import filterwarnings, warn

import numpy as np
from neuracore_types import (
    CameraData,
    Custom1DData,
    DataType,
    DepthCameraData,
    EndEffectorPoseData,
    JointData,
    LanguageData,
    ParallelGripperOpenAmountData,
    PointCloudData,
    PoseData,
    RGBCameraData,
)
from neuracore_types.utils import validate_safe_name

from neuracore.api.core import _get_robot
from neuracore.core.exceptions import RobotError
from neuracore.core.robot import Robot
from neuracore.core.streaming.data_stream import (
    DataStream,
    DepthDataStream,
    JsonDataStream,
    RGBDataStream,
    VideoDataStream,
)
from neuracore.core.streaming.p2p.stream_manager_orchestrator import (
    StreamManagerOrchestrator,
)
from neuracore.core.utils.depth_utils import MAX_DEPTH


class ExperimentalPointCloudWarning(UserWarning):
    """Warning for experimental point cloud features."""

    pass


filterwarnings("once", category=ExperimentalPointCloudWarning)


def start_stream(robot: Robot, data_stream: DataStream) -> None:
    """Start recording on a data stream if robot is currently recording.

    Args:
        robot: Robot instance
        data_stream: Data stream to start recording on
    """
    current_recording = robot.get_current_recording_id()
    if current_recording is not None and not data_stream.is_recording():
        data_stream.start_recording(current_recording)


def _log_single_joint_data(
    data_type: DataType, name: str, value: float, robot: Robot, timestamp: float
) -> None:
    """Log single joint data for a robot.

    Args:
        data_type: Type of joint data (e.g. DataType.JOINT_POSITIONS)
        name: Name of the joint
        value: Joint data value
        robot: Robot instance
        timestamp: Timestamp of the data
    """
    storage_name = validate_safe_name(name)
    str_id = f"{data_type.value}:{name}"
    joint_stream = robot.get_data_stream(str_id)
    if joint_stream is None:
        joint_stream = JsonDataStream(data_type=data_type, data_type_name=storage_name)
        robot.add_data_stream(str_id, joint_stream)

    start_stream(robot, joint_stream)

    data = JointData(
        timestamp=timestamp,
        value=value,
    )
    assert isinstance(
        joint_stream, JsonDataStream
    ), "Expected stream to be instance of JSONDataStream"
    joint_stream.log(data=data)
    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")
    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, data_type, sensor_key=str_id).publish(
        data.model_dump(mode="json")
    )


def _log_group_of_joint_data(
    data_type: DataType,
    joint_data: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint data for a robot.

    Args:
        data_type: Type of joint data (e.g. DataType.JOINT_POSITIONS)
        joint_data: Dictionary mapping joint names to joint data values
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If joint_data is not a dictionary of floats
    """
    timestamp = timestamp or time.time()
    if not isinstance(joint_data, dict):
        raise ValueError("Joint data must be a dictionary of floats")
    for key, value in joint_data.items():
        if not isinstance(value, float):
            raise ValueError(f"Joint data must be floats. {key} is not a float.")

    robot = _get_robot(robot_name, instance)
    for key, value in joint_data.items():
        _log_single_joint_data(data_type, key, value, robot, timestamp)


def _validate_extrinsics_intrinsics(
    extrinsics: np.ndarray | None, intrinsics: np.ndarray | None
) -> tuple[np.ndarray | None, np.ndarray | None]:
    """Validate and convert camera extrinsics and intrinsics matrices.

    Args:
        extrinsics: Optional extrinsics matrix as numpy array
        intrinsics: Optional intrinsics matrix as numpy array

    Returns:
        Tuple of validated (intrinsics, extrinsics) matrices

    Raises:
        ValueError: If matrices have incorrect shapes
    """
    if extrinsics is not None:
        if not isinstance(extrinsics, np.ndarray) or extrinsics.shape != (4, 4):
            raise ValueError("Extrinsics must be a numpy array of shape (4, 4)")
        extrinsics = extrinsics.astype(np.float16)
    if intrinsics is not None:
        if not isinstance(intrinsics, np.ndarray) or intrinsics.shape != (3, 3):
            raise ValueError("Intrinsics must be a numpy array of shape (3, 3)")
        intrinsics = intrinsics.astype(np.float16)
    return extrinsics, intrinsics


def _log_camera_data(
    camera_type: DataType,
    camera_data_without_frame: CameraData,
    image: np.ndarray,
    name: str,
    robot_name: str | None = None,
    instance: int = 0,
) -> None:
    """Log camera data for a robot.

    Args:
        camera_type: Type of camera (e.g. DataType.RGB or DataType.DEPTH)
        camera_data_without_frame: Camera data to log without frame
            (e.g. RGBCameraData or DepthCameraData)
        image: Image data as numpy array
        name: Unique identifier for the camera
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If image format is invalid or camera type is unsupported
    """
    assert camera_type in (
        DataType.RGB_IMAGES,
        DataType.DEPTH_IMAGES,
    ), "Unsupported camera type"

    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{camera_type.value}:{name}"

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    # data streaming for bucket storage (lossless and lossy)
    stream = robot.get_data_stream(str_id)
    # create the stream if it doesn't exist
    if stream is None:
        if camera_type == DataType.RGB_IMAGES:
            stream = RGBDataStream(storage_name, image.shape[1], image.shape[0])
        elif camera_type == DataType.DEPTH_IMAGES:
            stream = DepthDataStream(storage_name, image.shape[1], image.shape[0])
        else:
            raise ValueError(f"Invalid camera type: {camera_type}")
        robot.add_data_stream(str_id, stream)
    assert isinstance(
        stream, VideoDataStream
    ), "Expected stream as instance of VideoDataStream"

    start_stream(robot, stream)
    if stream.width != image.shape[1] or stream.height != image.shape[0]:
        raise ValueError(
            f"Camera image dimensions {image.shape[1]}x{image.shape[0]} do not match "
            f"stream dimensions {stream.width}x{stream.height}"
        )

    # NOTE: we explicitly do not include the frame in the
    # camera_data_without_frame object to avoid serializing the frame to JSON
    # or having to make two copies for streaming and bucket storage.
    camera_data_copy = camera_data_without_frame.model_copy()
    stream.log(camera_data_without_frame, frame=image)

    # peer to peer (p2p) streaming
    # NOTE: to avoid serializing the frame, we make another copy of the
    # camera_data_without_frame object because stream.log modifies the object
    # and adds the frame to it.
    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_video_source(name, camera_type, f"{name}_{camera_type}").add_frame(
        camera_data_copy, frame=image
    )


def log_custom_1d(
    name: str,
    data: np.ndarray,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log arbitrary data for a robot.

    Args:
        name: Name of the data stream
        data: Data to log (must be a numpy ndarray)
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If data is not JSON serializable
    """
    if not isinstance(data, np.ndarray):
        raise ValueError("Data must be a numpy ndarray")
    if data.ndim != 1:
        raise ValueError("Data must be a 1D numpy ndarray")
    timestamp = timestamp or time.time()
    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.CUSTOM_1D.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(
            data_type=DataType.CUSTOM_1D, data_type_name=storage_name
        )
        robot.add_data_stream(str_id, stream)

    start_stream(robot, stream)

    assert isinstance(
        stream, JsonDataStream
    ), "Expected stream to be instance of JSONDataStream"

    custom_data = Custom1DData(timestamp=timestamp, data=data)
    stream.log(custom_data)

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, DataType.CUSTOM_1D, sensor_key=str_id).publish(
        custom_data.model_dump(mode="json")
    )


def log_joint_positions(
    positions: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint positions for a robot.

    Args:
        positions: Dictionary mapping joint names to positions (in radians)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If positions is not a dictionary of floats
    """
    _log_group_of_joint_data(
        DataType.JOINT_POSITIONS,
        positions,
        robot_name,
        instance,
        timestamp,
    )


def log_joint_position(
    name: str,
    position: float,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint positions for a robot.

    Args:
        positions: Dictionary mapping joint names to positions (in radians)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If positions is not a dictionary of floats
    """
    _log_group_of_joint_data(
        DataType.JOINT_POSITIONS,
        {name: position},
        robot_name,
        instance,
        timestamp,
    )


def log_joint_target_positions(
    target_positions: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint target positions for a robot.

    Args:
        target_positions: Dictionary mapping joint names to
            target positions (in radians)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If target_positions is not a dictionary of floats
    """
    _log_group_of_joint_data(
        DataType.JOINT_TARGET_POSITIONS,
        target_positions,
        robot_name,
        instance,
        timestamp,
    )


def log_joint_target_position(
    name: str,
    target_position: float,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint target position for a robot.

    Args:
        name: Name of the joint
        target_position: Target position of the joint (in radians)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If target_position is not a float
    """
    _log_group_of_joint_data(
        DataType.JOINT_TARGET_POSITIONS,
        {name: target_position},
        robot_name,
        instance,
        timestamp,
    )


def log_joint_velocities(
    velocities: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint velocities for a robot.

    Args:
        velocities: Dictionary mapping joint names to velocities (in radians/second)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If velocities is not a dictionary of floats
    """
    _log_group_of_joint_data(
        DataType.JOINT_VELOCITIES,
        velocities,
        robot_name,
        instance,
        timestamp,
    )


def log_joint_velocity(
    name: str,
    velocity: float,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint velocity for a robot.

    Args:
        name: Name of the joint
        velocity: Velocity of the joint (in radians/second)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If velocity is not a float
    """
    _log_group_of_joint_data(
        DataType.JOINT_VELOCITIES,
        {name: velocity},
        robot_name,
        instance,
        timestamp,
    )


def log_joint_torques(
    torques: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint torques for a robot.

    Args:
        torques: Dictionary mapping joint names to torques (in Newton-meters)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If torques is not a dictionary of floats
    """
    _log_group_of_joint_data(
        DataType.JOINT_TORQUES,
        torques,
        robot_name,
        instance,
        timestamp,
    )


def log_joint_torque(
    name: str,
    torque: float,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log joint torque for a robot.

    Args:
        name: Name of the joint
        torque: Torque of the joint (in Newton-meters)
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If torque is not a float
    """
    _log_group_of_joint_data(
        DataType.JOINT_TORQUES,
        {name: torque},
        robot_name,
        instance,
        timestamp,
    )


def log_pose(
    name: str,
    pose: np.ndarray,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log pose data for a robot.

    Args:
        name: Name of the pose.
        pose: 7-element lists: [x, y, z, qx, qy, qz, qw]
        robot_name: Optional robot name.
            If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If poses is not a dictionary of 7-element lists
    """
    timestamp = timestamp or time.time()
    if not isinstance(pose, np.ndarray):
        raise ValueError(f"Poses must be lists. {name} is not a list.")
    if len(pose) != 7:
        raise ValueError(f"Poses must be lists of length 7. {name} is not length 7.")
    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.POSES.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(data_type=DataType.POSES, data_type_name=storage_name)
        robot.add_data_stream(str_id, stream)

    start_stream(robot, stream)
    assert isinstance(
        stream, JsonDataStream
    ), "Expected stream to be instance of JSONDataStream"

    pose_data = PoseData(timestamp=timestamp, pose=pose.tolist())
    stream.log(pose_data)

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, DataType.POSES, sensor_key=str_id).publish(
        pose_data.model_dump(mode="json")
    )


def log_end_effector_pose(
    name: str,
    pose: np.ndarray,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log end-effector pose data for a robot.

    Args:
        name: Name of the end effector
        pose: 7-element lists: [x, y, z, qx, qy, qz, qw]
        robot_name: Optional robot ID
        instance: Optional instance number
        timestamp: Optional timestamp
    """
    timestamp = timestamp or time.time()

    if not isinstance(pose, np.ndarray):
        raise ValueError(
            f"End effector pose must be a list. " f"{pose} is of type {type(pose)}"
        )
    if len(pose) != 7:
        raise ValueError(
            f"End effector pose must be a 7-element list. "
            f"{name} is of length {len(pose)}."
        )
    if not isinstance(name, str):
        raise ValueError(
            f"End effector names must be strings. " f"{name} is of type {type(name)}"
        )
    # check if last 4 elements of pose are a valid quaternion
    orientation = pose[3:]
    if not np.isclose(np.linalg.norm(orientation), 1.0, atol=1e-4):
        raise ValueError(
            f"End effector pose must be a valid unit quaternion. "
            f"{orientation} is not a valid unit quaternion."
        )

    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.END_EFFECTOR_POSES.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(
            data_type=DataType.END_EFFECTOR_POSES, data_type_name=storage_name
        )
        robot.add_data_stream(str_id, stream)

    start_stream(robot, stream)
    assert isinstance(stream, JsonDataStream)

    ee_pose_data = EndEffectorPoseData(timestamp=timestamp, pose=pose.tolist())
    stream.log(ee_pose_data)

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, DataType.END_EFFECTOR_POSES, sensor_key=str_id).publish(
        ee_pose_data.model_dump(mode="json")
    )


def log_parallel_gripper_open_amount(
    name: str,
    value: float,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log parallel gripper open amount data for a robot.

    Args:
        name: Name of the parallel gripper
        value: Open amount (0.0 = closed, 1.0 = fully open)
        robot_name: Optional robot ID
        instance: Optional instance number
        timestamp: Optional timestamp
    """
    timestamp = timestamp or time.time()
    if not isinstance(name, str):
        raise ValueError(
            f"Parallel gripper names must be strings. " f"{name} is not a string."
        )
    if not isinstance(value, float):
        raise ValueError(
            f"Parallel gripper open amounts must be floats. " f"{value} is not a float."
        )
    if value < 0.0 or value > 1.0:
        raise ValueError("Parallel gripper open amounts must be between 0.0 and 1.0.")

    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(
            data_type=DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS,
            data_type_name=storage_name,
        )
        robot.add_data_stream(str_id, stream)

    start_stream(robot, stream)
    assert isinstance(stream, JsonDataStream)

    parallel_gripper_open_amount_data = ParallelGripperOpenAmountData(
        timestamp=timestamp, open_amount=value
    )
    stream.log(parallel_gripper_open_amount_data)

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS, str_id).publish(
        parallel_gripper_open_amount_data.model_dump(mode="json")
    )


def log_parallel_gripper_open_amounts(
    values: dict[str, float],
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log parallel gripper open amount data for a robot.

    Args:
        values: Dictionary mapping gripper names to open amounts
            (0.0 = closed, 1.0 = fully open)
        robot_name: Optional robot ID
        instance: Optional instance number
        timestamp: Optional timestamp
    """
    timestamp = timestamp or time.time()
    for name, value in values.items():
        log_parallel_gripper_open_amount(
            name=name,
            value=value,
            robot_name=robot_name,
            instance=instance,
            timestamp=timestamp,
        )


def log_language(
    name: str,
    language: str,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log language annotation for a robot.

    Args:
        name: Name of the language annotation
        language: A language string associated with this timestep
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If language is not a string
    """
    timestamp = timestamp or time.time()
    if not isinstance(language, str):
        raise ValueError("Language must be a string")
    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.LANGUAGE.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(
            data_type=DataType.LANGUAGE, data_type_name=storage_name
        )
        robot.add_data_stream(str_id, stream)
    start_stream(robot, stream)
    assert isinstance(
        stream, JsonDataStream
    ), "Expected stream to be instance of JSONDataStream"

    data = LanguageData(timestamp=timestamp, text=language)
    stream.log(data)

    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    StreamManagerOrchestrator().get_provider_manager(
        robot.id, robot.instance
    ).get_json_source(str_id, DataType.LANGUAGE, sensor_key=str_id).publish(
        data.model_dump(mode="json")
    )


def log_rgb(
    name: str,
    rgb: np.ndarray,
    extrinsics: np.ndarray | None = None,
    intrinsics: np.ndarray | None = None,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log RGB image from a camera.

    Args:
        name: Unique identifier for the camera
        rgb: RGB image as numpy array (HxWx3, dtype=uint8)
        extrinsics: Optional extrinsics matrix (4x4)
        intrinsics: Optional intrinsics matrix (3x3)
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If image format is invalid
    """
    if not isinstance(rgb, np.ndarray):
        raise ValueError("Image image must be a numpy array")
    if rgb.dtype != np.uint8:
        raise ValueError("Image must be uint8 with range 0-255")
    extrinsics, intrinsics = _validate_extrinsics_intrinsics(extrinsics, intrinsics)
    timestamp = timestamp or time.time()
    rgb_camera_data = RGBCameraData(
        timestamp=timestamp,
        extrinsics=extrinsics,
        intrinsics=intrinsics,
        frame=None,
    )
    _log_camera_data(
        DataType.RGB_IMAGES,
        rgb_camera_data,
        rgb,
        name,
        robot_name,
        instance,
    )


def log_depth(
    name: str,
    depth: np.ndarray,
    extrinsics: np.ndarray | None = None,
    intrinsics: np.ndarray | None = None,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log depth image from a camera.

    Args:
        name: Unique identifier for the camera
        depth: Depth image as numpy array (HxW, dtype=float16 or float32, in meters)
        extrinsics: Optional extrinsics matrix (4x4)
        intrinsics: Optional intrinsics matrix (3x3)
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If depth format is invalid
    """
    if not isinstance(depth, np.ndarray):
        raise ValueError("Depth image must be a numpy array")
    if depth.dtype not in (np.float16, np.float32):
        raise ValueError(
            f"Depth image must be float16 or float32, but got {depth.dtype}"
        )
    if depth.max() > MAX_DEPTH:
        raise ValueError(
            "Depth image should be in meters. "
            f"You are attempting to log depth values > {MAX_DEPTH}. "
            "The values you are passing in are likely in millimeters."
        )
    extrinsics, intrinsics = _validate_extrinsics_intrinsics(extrinsics, intrinsics)
    timestamp = timestamp or time.time()
    depth_camera_data = DepthCameraData(
        timestamp=timestamp,
        extrinsics=extrinsics,
        intrinsics=intrinsics,
        frame=None,
    )
    _log_camera_data(
        DataType.DEPTH_IMAGES,
        depth_camera_data,
        depth,
        name,
        robot_name,
        instance,
    )


def log_point_cloud(
    name: str,
    points: np.ndarray,
    rgb_points: np.ndarray | None = None,
    extrinsics: np.ndarray | None = None,
    intrinsics: np.ndarray | None = None,
    robot_name: str | None = None,
    instance: int = 0,
    timestamp: float | None = None,
) -> None:
    """Log point cloud data from a camera.

    Args:
        name: Unique identifier for the point cloud
        points: Point cloud as numpy array (Nx3, dtype=float32, in meters)
        rgb_points: Optional RGB values for each point (Nx3, dtype=uint8)
        extrinsics: Optional extrinsics matrix (4x4)
        intrinsics: Optional intrinsics matrix (3x3)
        robot_name: Optional robot ID. If not provided, uses the last initialized robot
        instance: Optional instance number of the robot
        timestamp: Optional timestamp

    Raises:
        RobotError: If no robot is active and no robot_name provided
        ValueError: If point cloud format is invalid
    """
    warn(
        "Point cloud logging is experimental and may change in future releases.",
        ExperimentalPointCloudWarning,
    )
    timestamp = timestamp or time.time()
    if not isinstance(points, np.ndarray):
        raise ValueError("Point cloud must be a numpy array")
    if points.dtype != np.float16:
        raise ValueError("Point cloud must be float16")
    if points.shape[1] != 3:
        raise ValueError("Point cloud must have 3 columns")
    if points.shape[0] > 307200:
        raise ValueError("Point cloud must have at most 307200 points")

    if rgb_points is not None:
        if not isinstance(rgb_points, np.ndarray):
            raise ValueError("RGB point cloud must be a numpy array")
        if rgb_points.dtype != np.uint8:
            raise ValueError("RGB point cloud must be uint8")
        if rgb_points.shape[0] != points.shape[0]:
            raise ValueError(
                "RGB point cloud must have the same number of points as the point cloud"
            )
        if rgb_points.shape[1] != 3:
            raise ValueError("RGB point cloud must have 3 columns")

    extrinsics, intrinsics = _validate_extrinsics_intrinsics(extrinsics, intrinsics)
    robot = _get_robot(robot_name, instance)
    storage_name = validate_safe_name(name)
    str_id = f"{DataType.POINT_CLOUDS.value}:{name}"
    stream = robot.get_data_stream(str_id)
    if stream is None:
        stream = JsonDataStream(
            data_type=DataType.POINT_CLOUDS, data_type_name=storage_name
        )
        robot.add_data_stream(str_id, stream)
    assert isinstance(
        stream, JsonDataStream
    ), "Expected stream to be instance of JSONDataStream"
    start_stream(robot, stream)
    point_data = PointCloudData(
        timestamp=timestamp,
        points=points,
        rgb_points=rgb_points,
        extrinsics=extrinsics,
        intrinsics=intrinsics,
    )
    stream.log(point_data)
    if robot.id is None:
        raise RobotError("Robot not initialized. Call init() first.")

    json_data = point_data.model_dump(mode="json")
    src = (
        StreamManagerOrchestrator()
        .get_provider_manager(robot.id, robot.instance)
        .get_json_source(str_id, DataType.POINT_CLOUDS, sensor_key=str_id)
    )
    src.publish(json_data)



"""Dummy dataset for algorithm validation and testing without real data.

This module provides a synthetic dataset that generates random data matching
the structure of real Neuracore datasets. It's used for algorithm development,
testing, and validation without requiring actual robot demonstration data.
"""

import copy
import logging

import torch
from neuracore_types import (
    DATA_TYPE_TO_BATCHED_NC_DATA_CLASS,
    DATA_TYPE_TO_NC_DATA_CLASS,
    BatchedNCData,
    DataType,
    NCData,
    NCDataStats,
    RobotDataSpec,
)

from neuracore.core.robot import Robot
from neuracore.ml import BatchedTrainingSamples
from neuracore.ml.datasets.pytorch_neuracore_dataset import PytorchNeuracoreDataset

logger = logging.getLogger(__name__)

TrainingSample = BatchedTrainingSamples

T = 1
MAX_LEN_PER_DATA_TYPE = 2


class PytorchDummyDataset(PytorchNeuracoreDataset):
    """Synthetic dataset for algorithm validation and testing.

    This dataset generates random data with the same structure and dimensions
    as real Neuracore datasets, allowing for algorithm development and testing
    without requiring actual robot demonstration data. It supports all standard
    data types including images, joint data, depth images, point clouds,
    poses, end-effectors, and language instructions.
    """

    def __init__(
        self,
        input_robot_data_spec: RobotDataSpec,
        output_robot_data_spec: RobotDataSpec,
        num_samples: int = 50,
        num_episodes: int = 10,
        output_prediction_horizon: int = 5,
    ):
        """Initialize the dummy dataset with specified data types and dimensions.

        Args:
            input_robot_data_spec: Mapping from robot_id
                to data spec for model inputs.
            output_robot_data_spec: Mapping from robot_id
                to data spec for model outputs.
            num_samples: Total number of training samples to generate.
            num_episodes: Number of distinct episodes in the dataset.
            output_prediction_horizon: Length of output action sequences.
        """
        super().__init__(
            num_recordings=num_episodes,
            input_robot_data_spec=input_robot_data_spec,
            output_robot_data_spec=output_robot_data_spec,
            output_prediction_horizon=output_prediction_horizon,
        )
        self.num_samples = num_samples
        self.robot = Robot("dummy_robot", 0)
        self.robot.id = "dummy_robot_id"

        self._dataset_statistics: dict[DataType, list[NCDataStats]] = {}

        # Generate sample data once and reuse for all samples
        self._sample_to_return = self._generate_sample()

    def _generate_sample(self) -> TrainingSample:
        """Generate a single training sample template with random data.

        Creates synthetic data for all specified input and output data types,
        with appropriate dimensions and masking.

        Returns:
            A TrainingSample containing randomly generated input and output data.
        """
        inputs_and_outputs: dict[DataType, list[BatchedNCData]] = {}
        inputs_and_outputs_mask: dict[DataType, torch.Tensor] = {}

        # Generate data for all data types in the merged spec
        # robot_data_spec is dict[robot_id, dict[DataType, list[str]]]
        for robot_id, data_spec in self.robot_data_spec.items():
            for data_type, data_names in data_spec.items():
                num_items = (
                    len(data_names) if len(data_names) > 0 else MAX_LEN_PER_DATA_TYPE
                )

                if data_type not in inputs_and_outputs:
                    inputs_and_outputs[data_type] = []
                    inputs_and_outputs_mask[data_type] = torch.ones(
                        (num_items,), dtype=torch.float32
                    )
                    self._dataset_statistics[data_type] = []

                # Generate data for each item (e.g., each joint, each camera)
                for _ in range(num_items):
                    nc_data_class = DATA_TYPE_TO_NC_DATA_CLASS[data_type]
                    sampled_data: NCData = nc_data_class.sample()

                    # Store statistics for this data item
                    self._dataset_statistics[data_type].append(
                        sampled_data.calculate_statistics()
                    )

                    # Generate batched data
                    batched_nc_data_type = DATA_TYPE_TO_BATCHED_NC_DATA_CLASS[data_type]
                    batched_nc_data = batched_nc_data_type.sample(
                        batch_size=1, time_steps=T
                    )

                    # Drop batch dimension for storage
                    for attr_name in vars(batched_nc_data):
                        attr_value = getattr(batched_nc_data, attr_name)
                        if (
                            isinstance(attr_value, torch.Tensor)
                            and attr_value.shape[0] == 1
                        ):
                            setattr(batched_nc_data, attr_name, attr_value)

                    inputs_and_outputs[data_type].append(batched_nc_data)

        # Separate inputs and outputs based on specs
        inputs: dict[DataType, list[BatchedNCData]] = {}
        inputs_mask: dict[DataType, torch.Tensor] = {}
        outputs: dict[DataType, list[BatchedNCData]] = {}
        outputs_mask: dict[DataType, torch.Tensor] = {}

        # Collect input data types from all robots
        for robot_id, data_spec in self.input_robot_data_spec.items():
            for data_type in data_spec.keys():
                if data_type not in inputs:
                    inputs[data_type] = inputs_and_outputs[data_type]
                    inputs_mask[data_type] = inputs_and_outputs_mask[data_type]

        # Collect output data types from all robots
        for robot_id, data_spec in self.output_robot_data_spec.items():
            for data_type in data_spec.keys():
                if data_type not in outputs:
                    # Extend output temporal dimension by prediction horizon
                    outputs[data_type] = []
                    for output in inputs_and_outputs[data_type]:
                        output = copy.deepcopy(output)
                        for attr_name in vars(output):
                            attr_value = getattr(output, attr_name)
                            if isinstance(attr_value, torch.Tensor):
                                # Repeat along time dimension for prediction horizon
                                attr_value = attr_value.repeat_interleave(
                                    self.output_prediction_horizon, dim=1
                                )
                                setattr(output, attr_name, attr_value)
                        outputs[data_type].append(output)
                    outputs_mask[data_type] = inputs_and_outputs_mask[data_type]

        return TrainingSample(
            inputs=inputs,
            inputs_mask=inputs_mask,
            outputs=outputs,
            outputs_mask=outputs_mask,
            batch_size=0,
        )

    def load_sample(
        self, episode_idx: int, timestep: int | None = None
    ) -> TrainingSample:
        """Generate a random training sample with realistic data structure.

        Creates synthetic data that matches the format and dimensions of real
        robot demonstration data, including appropriate masking and tensor shapes.

        Args:
            episode_idx: Index of the episode (used for reproducible randomness).
            timestep: Optional timestep within the episode (currently unused).

        Returns:
            A TrainingSample containing randomly generated input and output data
            matching the specified data types and dimensions.
        """
        return self._sample_to_return

    def __len__(self) -> int:
        """Get the total number of samples in the dataset.

        Returns:
            The number of training samples available in this dataset.
        """
        return self.num_samples

    def __getitem__(self, idx: int) -> BatchedTrainingSamples:
        """Get a training sample from the dataset by index.

        Args:
            idx: Index of the sample to retrieve.

        Returns:
            A TrainingSample containing the requested data.
        """
        if idx < 0 or idx >= self.num_samples:
            raise IndexError("Index out of range for dataset.")
        return self.load_sample(0, 0)

    @property
    def dataset_statistics(self) -> dict[DataType, list[NCDataStats]]:
        """Return the dataset description.

        Returns:
            A dictionary mapping each DataType to a list of NCDataStats
            describing the statistics of that data type in the dataset.
        """
        return self._dataset_statistics



"""Model archive creation utility for Neuracore model deployment.

This module provides functionality to package Neuracore models into simple
ZIP archives (.nc.zip) for deployment. It handles model serialization,
dependency management, and packaging of all required files for inference.
"""

import inspect
import json
import logging
import tempfile
import zipfile
from pathlib import Path
from typing import Any

import torch
from neuracore_types import ModelInitDescription

from neuracore.ml.core.neuracore_model import NeuracoreModel
from neuracore.ml.utils.algorithm_loader import AlgorithmLoader
from neuracore.ml.utils.device_utils import get_default_device

logger = logging.getLogger(__name__)


def create_nc_archive(
    model: NeuracoreModel, output_dir: Path, algorithm_config: dict = {}
) -> Path:
    """Create a Neuracore model archive (NC.ZIP) file from a Neuracore model.

    Packages a trained Neuracore model into a deployable ZIP file that includes
    the model weights, algorithm code, configuration metadata, and dependencies.
    The resulting NC.ZIP file can be deployed for inference.

    Args:
        model: Trained Neuracore model instance to package for deployment.
        output_dir: Directory path where the NC.ZIP file will be created.
        algorithm_config: Custom configuration for the algorithm.

    Returns:
        Path to the created NC.ZIP file.
    """
    algorithm_file = Path(inspect.getfile(model.__class__))
    algorithm_loader = AlgorithmLoader(algorithm_file.parent)
    algo_files = algorithm_loader.get_all_files()
    requirements_file_path = algorithm_loader.algorithm_dir / "requirements.txt"

    # Create output directory if it doesn't exist
    output_dir.mkdir(parents=True, exist_ok=True)

    # Create the archive filename
    archive_path = output_dir / "model.nc.zip"

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_path = Path(temp_dir)

        # Save model weights
        torch.save(model.state_dict(), temp_path / "model.pt")
        model_size_mb = (temp_path / "model.pt").stat().st_size / (1024 * 1024)
        logger.info("Model weights saved (%.1f MB)", model_size_mb)

        # Save model initialization description
        with open(temp_path / "model_init_description.json", "w") as f:
            json.dump(model.model_init_description.model_dump(mode="json"), f, indent=2)

        # Save algorithm config (always create file, even if empty)
        with open(temp_path / "algorithm_config.json", "w") as f:
            json.dump(algorithm_config, f, indent=2)

        # Create the ZIP archive
        with zipfile.ZipFile(archive_path, "w", zipfile.ZIP_DEFLATED) as zip_file:
            zip_file.write(
                temp_path / "model.pt", "model.pt", compress_type=zipfile.ZIP_STORED
            )

            # Add model initialization description
            zip_file.write(
                temp_path / "model_init_description.json", "model_init_description.json"
            )

            # Add algorithm config (always present)
            zip_file.write(temp_path / "algorithm_config.json", "algorithm_config.json")

            # Add all algorithm files
            for algo_file in algo_files:
                # Calculate relative path from algorithm directory
                rel_path = algo_file.relative_to(algorithm_loader.algorithm_dir)
                zip_file.write(algo_file, f"algorithm/{rel_path}")

            # Add requirements file if it exists
            if requirements_file_path.exists():
                zip_file.write(requirements_file_path, "requirements.txt")

    archive_size_mb = archive_path.stat().st_size / (1024 * 1024)
    logger.info(
        "NC archive created successfully: %s (%.1f MB)", archive_path, archive_size_mb
    )
    return archive_path


def extract_nc_archive(archive_file: Path, output_dir: Path) -> dict[str, Path]:
    """Extract all contents from a Neuracore model archive (NC.ZIP) file.

    Extracts all files from a NC.ZIP archive including model weights, algorithm code,
    configuration files, and dependencies.

    Args:
        archive_file: Path to the NC.ZIP file to extract.
        output_dir: Directory where extracted files will be saved.

    Returns:
        Dictionary mapping file types to their extracted paths.

    Raises:
        FileNotFoundError: If the archive file doesn't exist.
        zipfile.BadZipFile: If the archive file is corrupted or not a valid ZIP.
    """
    if not archive_file.exists():
        raise FileNotFoundError(f"Archive file not found: {archive_file}")

    output_dir.mkdir(parents=True, exist_ok=True)
    extracted_files: dict[str, Any] = {}

    with zipfile.ZipFile(archive_file, "r") as zip_ref:
        # Extract all files
        zip_ref.extractall(output_dir)

        # Catalog the extracted files
        for file_info in zip_ref.filelist:
            file_path = output_dir / file_info.filename

            # Categorize files based on their names/extensions
            if file_info.filename == "model.pt":
                extracted_files["model_weights"] = file_path
            elif file_info.filename == "model_init_description.json":
                extracted_files["model_init_description"] = file_path
            elif file_info.filename == "algorithm_config.json":
                extracted_files["algorithm_config"] = file_path
            elif file_info.filename == "requirements.txt":
                extracted_files["requirements"] = file_path
            elif file_info.filename.startswith("algorithm/"):
                extracted_files.setdefault("algorithm_files", []).append(file_path)
            else:
                extracted_files.setdefault("other_files", []).append(file_path)

    return extracted_files


def load_model_from_nc_archive(
    archive_file: Path, extract_to: Path | None = None, device: str | None = None
) -> NeuracoreModel:
    """Load a Neuracore model from a NC.ZIP archive file.

    Extracts the archive file and reconstructs the original Neuracore model instance
    with its trained weights and configuration.

    Args:
        archive_file: Path to the NC.ZIP file.
        extract_to: Optional directory to extract files to.
            If None, uses a temporary directory.
        device: Optional device model to be loaded on

    Returns:
        NeuracoreModel: The reconstructed model instance ready for inference.
    """
    use_temp_dir = extract_to is None

    if use_temp_dir:
        temp_dir_context = tempfile.TemporaryDirectory()
        extract_to = Path(temp_dir_context.__enter__())
    else:
        temp_dir_context = None

    assert extract_to is not None

    try:
        # Extract the archive file
        extracted_files = extract_nc_archive(archive_file, extract_to)

        # Load model initialization description
        if "model_init_description" not in extracted_files:
            raise FileNotFoundError("model_init_description.json not found in archive")

        with open(extracted_files["model_init_description"]) as f:
            model_init_description = json.load(f)
        model_init_description = ModelInitDescription.model_validate(
            model_init_description
        )

        # Load algorithm config if present
        algorithm_config = {}
        if "algorithm_config" in extracted_files:
            with open(extracted_files["algorithm_config"]) as f:
                algorithm_config = json.load(f)

        # Find the algorithm directory
        algorithm_dir = extract_to / "algorithm"
        if not algorithm_dir.exists():
            raise FileNotFoundError("Algorithm directory not found in archive")

        # Load the algorithm using AlgorithmLoader
        algorithm_loader = AlgorithmLoader(algorithm_dir)
        algorithm_loader.install_requirements()
        model_class = algorithm_loader.load_model()

        # Create model instance
        if device:
            device = torch.device(device)
        else:
            device = get_default_device()
        model = model_class(model_init_description, **algorithm_config)
        model = model.to(device)

        # Load trained weights if present
        if "model_weights" in extracted_files:
            state_dict = torch.load(
                extracted_files["model_weights"],
                map_location=model.device,
                weights_only=True,
            )
            model.load_state_dict(state_dict)

        return model

    finally:
        if use_temp_dir and temp_dir_context:
            temp_dir_context.__exit__(None, None, None)



"""MJCF to URDF conversion utility for robot model formats.

This module provides functionality to convert MuJoCo XML format (MJCF) robot
models to Universal Robot Description Format (URDF). It handles the conversion
of kinematic structures, inertial properties, joint configurations, and mesh
geometries while creating appropriate STL files for visual elements.
"""

from pathlib import Path
from xml.dom import minidom
from xml.etree import ElementTree as ET

import mujoco
import numpy as np
from scipy.spatial.transform import Rotation
from stl import mesh


def _array2str(arr: list) -> str:
    """Convert numeric array to space-separated string for XML attributes.

    Args:
        arr: List or array of numeric values.

    Returns:
        Space-separated string representation of the array values.
    """
    return " ".join([str(x) for x in arr])


def _create_body(
    xml_root: ET.Element,
    name: str,
    inertial_pos: list,
    inertial_rpy: list,
    mass: float,
    ixx: float,
    iyy: float,
    izz: float,
) -> ET.Element:
    """Create a URDF link element with specified mass and inertia properties.

    Args:
        xml_root: Parent XML element to attach the new link to.
        name: Name identifier for the link.
        inertial_pos: Position offset of the center of mass [x, y, z].
        inertial_rpy: Orientation of the inertial frame in roll-pitch-yaw [r, p, y].
        mass: Mass of the link in kilograms.
        ixx: Moment of inertia about the x-axis.
        iyy: Moment of inertia about the y-axis.
        izz: Moment of inertia about the z-axis.

    Returns:
        The created link XML element.
    """
    # create XML element for this body
    body = ET.SubElement(xml_root, "link", {"name": name})

    # add inertial element
    inertial = ET.SubElement(body, "inertial")
    ET.SubElement(
        inertial,
        "origin",
        {"xyz": _array2str(inertial_pos), "rpy": _array2str(inertial_rpy)},
    )
    ET.SubElement(inertial, "mass", {"value": str(mass)})
    ET.SubElement(
        inertial,
        "inertia",
        {
            "ixx": str(ixx),
            "iyy": str(iyy),
            "izz": str(izz),
            "ixy": "0",
            "ixz": "0",
            "iyz": "0",
        },
    )
    return body


def _create_dummy_body(xml_root: ET.Element, name: str) -> ET.Element:
    """Create a lightweight dummy link with negligible mass and inertia.

    Used for creating intermediate joint bodies in the URDF kinematic chain
    where MJCF and URDF joint representations differ.

    Args:
        xml_root: Parent XML element to attach the new link to.
        name: Name identifier for the dummy link.

    Returns:
        The created dummy link XML element.
    """
    mass = 0.001
    mass_moi = mass * (0.001**2)  # mass moment of inertia
    return _create_body(
        xml_root, name, np.zeros(3), np.zeros(3), mass, mass_moi, mass_moi, mass_moi
    )


def _create_joint(
    xml_root: ET.Element,
    name: str,
    parent: str,
    child: str,
    pos: list,
    rpy: list,
    axis: list | None = None,
    jnt_range: list | None = None,
    jnt_type: str = "fixed",
) -> ET.Element:
    """Create a URDF joint element connecting two links.

    Supports fixed, revolute, prismatic, continuous, and floating joints.
    For floating joints, no axis or limits are added (6-DoF free joint).

    Args:
        xml_root: Parent XML element to attach the new joint to.
        name: Name identifier for the joint.
        parent: Name of the parent link.
        child: Name of the child link.
        pos: Position offset from parent to child [x, y, z].
        rpy: Orientation offset in roll-pitch-yaw [r, p, y].
        axis: Joint axis of rotation/translation [x, y, z].
        jnt_range: Joint limits [min, max].
        jnt_type: Type of joint ("fixed", "revolute", "prismatic",
                  "continuous", or "floating").

    Returns:
        The created joint XML element.
    """
    # Basic validation of parameters vs joint type
    if jnt_type in ("fixed", "floating"):
        # These joint types should not have an axis or limits
        if axis is not None or jnt_range is not None:
            raise ValueError(
                f"Joint type '{jnt_type}' should not have axis or limits "
                f"(got axis={axis}, range={jnt_range})."
            )
    elif jnt_type in ("revolute", "continuous", "prismatic"):
        # These DOF joints must have an axis
        if axis is None:
            raise ValueError(f"Joint type '{jnt_type}' requires an axis.")
    else:
        raise ValueError(f"Unsupported joint type: {jnt_type}")

    # Create joint element
    jnt_element = ET.SubElement(xml_root, "joint", {"type": jnt_type, "name": name})
    ET.SubElement(jnt_element, "parent", {"link": parent})
    ET.SubElement(jnt_element, "child", {"link": child})
    ET.SubElement(
        jnt_element, "origin", {"xyz": _array2str(pos), "rpy": _array2str(rpy)}
    )

    if axis is None:
        axis = [0, 0, 0]

    # Only non-floating / non-fixed joints get axis and limits
    if jnt_type in ("revolute", "continuous", "prismatic"):
        ET.SubElement(jnt_element, "axis", {"xyz": _array2str(axis)})
        if jnt_range is not None:
            ET.SubElement(
                jnt_element,
                "limit",
                {
                    "lower": str(jnt_range[0]),
                    "upper": str(jnt_range[1]),
                    "effort": "100",
                    "velocity": "100",
                },
            )

    return jnt_element


def convert(mjcf_file: str, urdf_file: Path, asset_file_prefix: str = "") -> None:
    """Convert a MuJoCo MJCF file to URDF format.

    Performs a comprehensive conversion from MJCF to URDF including:
    - Kinematic structure and joint relationships
    - Mass and inertia properties
    - Joint types and limits (revolute, prismatic, fixed)
    - Visual mesh geometries converted to STL format
    - Proper coordinate frame transformations

    The conversion handles the structural differences between MJCF and URDF
    by creating intermediate joint bodies when necessary to maintain equivalent
    kinematic behavior.

    Args:
        mjcf_file: Path to the input MJCF file to convert.
        urdf_file: Path where the output URDF file will be saved.
        asset_file_prefix: Optional prefix for generated STL mesh files.
            Useful for organizing assets in subdirectories.

    Raises:
        AssertionError: If a body has more than one joint, which is not supported.
    """
    model = mujoco.MjModel.from_xml_path(mjcf_file)
    root = ET.Element("robot", {"name": "converted_robot"})

    for id in range(model.nbody):
        child_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_BODY, id)
        parent_id = model.body_parentid[id]
        parent_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_BODY, parent_id)

        parentbody2childbody_pos = model.body_pos[id]
        parentbody2childbody_quat = model.body_quat[id]  # [w, x, y, z]
        # change to [x, y, z, w]
        parentbody2childbody_quat = [
            parentbody2childbody_quat[1],
            parentbody2childbody_quat[2],
            parentbody2childbody_quat[3],
            parentbody2childbody_quat[0],
        ]
        parentbody2childbody_Rot = Rotation.from_quat(
            parentbody2childbody_quat
        ).as_matrix()
        parentbody2childbody_rpy = Rotation.from_matrix(
            parentbody2childbody_Rot
        ).as_euler("xyz")

        # read inertial info
        mass = model.body_mass[id]
        inertia = model.body_inertia[id]

        childbody2childinertia_pos = model.body_ipos[id]
        childbody2childinertia_quat = model.body_iquat[id]  # [w, x, y, z]
        # change to [x, y, z, w]
        childbody2childinertia_quat = [
            childbody2childinertia_quat[1],
            childbody2childinertia_quat[2],
            childbody2childinertia_quat[3],
            childbody2childinertia_quat[0],
        ]
        childbody2childinertia_Rot = Rotation.from_quat(
            childbody2childinertia_quat
        ).as_matrix()
        childbody2childinertia_rpy = Rotation.from_matrix(
            childbody2childinertia_Rot
        ).as_euler("xyz")

        jntnum = model.body_jntnum[id]
        assert jntnum <= 1, "only one joint per body supported"

        if jntnum == 1:
            # load joint info
            jntid = model.body_jntadr[id]
            jnt_type = model.jnt_type[jntid]

            jnt_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_JOINT, jntid)

            if jnt_name is None:
                jnt_name = f"{parent_name}2{child_name}_joint"

            jnt_range = None
            jnt_axis_childbody = None
            childbody2jnt_pos = model.jnt_pos[jntid]  # [x, y, z]

            # Handle different joint types
            if jnt_type == mujoco.mjtJoint.mjJNT_HINGE:
                urdf_jnt_type = "revolute"
                jnt_range = model.jnt_range[jntid]  # [min, max]
                jnt_axis_childbody = model.jnt_axis[jntid]  # [x, y, z]
            elif jnt_type == mujoco.mjtJoint.mjJNT_SLIDE:
                urdf_jnt_type = "prismatic"
                jnt_range = model.jnt_range[jntid]  # [min, max]
                jnt_axis_childbody = model.jnt_axis[jntid]  # [x, y, z]
            elif jnt_type == mujoco.mjtJoint.mjJNT_FREE:
                urdf_jnt_type = "floating"
            else:
                # raise a value error for unsupported joint types
                raise ValueError(f"unsupported joint type: {jnt_type}")

            parentbody2jnt_axis = jnt_axis_childbody
        else:
            # create a fixed joint instead
            jnt_name = f"{parent_name}2{child_name}_fixed"
            urdf_jnt_type = "fixed"
            jnt_range = None
            childbody2jnt_pos = np.zeros(3)
            parentbody2jnt_axis = None

        # create child body
        body_element = _create_body(
            root,
            child_name,
            childbody2childinertia_pos,
            childbody2childinertia_rpy,
            mass,
            inertia[0],
            inertia[1],
            inertia[2],
        )

        # read geom info and add it child body
        geomnum = model.body_geomnum[id]
        for geomnum_i in range(geomnum):
            geomid = model.body_geomadr[id] + geomnum_i
            if model.geom_type[geomid] != mujoco.mjtGeom.mjGEOM_MESH:
                # only support mesh geoms
                continue
            geom_dataid = model.geom_dataid[geomid]  # id of geom's mesh
            geom_pos = model.geom_pos[geomid]
            geom_quat = model.geom_quat[geomid]  # [w, x, y, z]
            # change to [x, y, z, w]
            geom_quat = [geom_quat[1], geom_quat[2], geom_quat[3], geom_quat[0]]
            geom_rpy = Rotation.from_quat(geom_quat).as_euler("xyz")
            mesh_name = mujoco.mj_id2name(model, mujoco.mjtObj.mjOBJ_MESH, geom_dataid)
            # create visual element within body element
            visual_element = ET.SubElement(body_element, "visual", {"name": mesh_name})
            ET.SubElement(
                visual_element,
                "origin",
                {"xyz": _array2str(geom_pos), "rpy": _array2str(geom_rpy)},
            )
            geometry_element = ET.SubElement(visual_element, "geometry")
            ET.SubElement(
                geometry_element,
                "mesh",
                {"filename": f"{asset_file_prefix}converted_{mesh_name}.stl"},
            )
            material_element = ET.SubElement(
                visual_element, "material", {"name": "white"}
            )

            # create STL
            vertadr = model.mesh_vertadr[geom_dataid]
            vertnum = model.mesh_vertnum[geom_dataid]
            vert = model.mesh_vert[vertadr : vertadr + vertnum]
            faceadr = model.mesh_faceadr[geom_dataid]
            facenum = model.mesh_facenum[geom_dataid]
            face = model.mesh_face[faceadr : faceadr + facenum]
            data = np.zeros(facenum, dtype=mesh.Mesh.dtype)
            for i in range(facenum):
                data["vectors"][i] = vert[face[i]]
            m = mesh.Mesh(data, remove_empty_areas=False)
            urdf_base = urdf_file.parent
            if asset_file_prefix:
                stl_file = urdf_base / asset_file_prefix / f"converted_{mesh_name}.stl"
            else:
                stl_file = urdf_base / f"converted_{mesh_name}.stl"
            stl_file.parent.mkdir(parents=True, exist_ok=True)
            m.save(stl_file)

        if child_name == "world":
            # there is no joint connecting the world to anything, since it is the root
            assert parent_name == "world"
            assert jntnum == 0
            continue  # skip adding joint element or parent body

        # create dummy body for joint
        jnt_body_name = f"{jnt_name}_jointbody"
        _create_dummy_body(root, jnt_body_name)
        # connect parent to joint body with appropriate joint type
        parentbody2jnt_pos = (
            parentbody2childbody_pos + parentbody2childbody_Rot @ childbody2jnt_pos
        )
        parentbody2jnt_rpy = parentbody2childbody_rpy
        _create_joint(
            root,
            jnt_name,
            parent_name,
            jnt_body_name,
            parentbody2jnt_pos,
            parentbody2jnt_rpy,
            parentbody2jnt_axis,
            jnt_range,
            urdf_jnt_type,
        )
        # connect joint body to child body with fixed joint
        jnt2childbody_pos = -childbody2jnt_pos
        jnt2childbody_rpy = np.zeros(3)
        _create_joint(
            root,
            f"{jnt_name}_offset",
            jnt_body_name,
            child_name,
            jnt2childbody_pos,
            jnt2childbody_rpy,
        )

    # define white material
    material_element = ET.SubElement(root, "material", {"name": "white"})
    ET.SubElement(material_element, "color", {"rgba": "1 1 1 1"})

    # write to file with pretty printing
    xmlstr = minidom.parseString(ET.tostring(root)).toprettyxml(indent="   ")
    with urdf_file.open("w") as f:
        f.write(xmlstr)



"""CNN+MLP model for robot manipulation with sequence prediction.

This module implements a simple baseline model that combines convolutional
neural networks for visual feature extraction with multi-layer perceptrons
for action sequence prediction. The model processes single timestep inputs
and outputs entire action sequences.
"""

import os
from typing import Any, cast

import torch
import torch.nn as nn
import torchvision.transforms as T
from neuracore_types import (
    BatchedDepthData,
    BatchedJointData,
    BatchedLanguageData,
    BatchedNCData,
    BatchedParallelGripperOpenAmountData,
    BatchedPointCloudData,
    BatchedPoseData,
    BatchedRGBData,
    CameraDataStats,
    DataItemStats,
    DataType,
    JointDataStats,
    ModelInitDescription,
    ParallelGripperOpenAmountDataStats,
    PointCloudDataStats,
)

from neuracore.ml import (
    BatchedInferenceInputs,
    BatchedTrainingOutputs,
    BatchedTrainingSamples,
    NeuracoreModel,
)
from neuracore.ml.algorithm_utils.modules import (
    DepthImageEncoder,
    MultimodalFusionEncoder,
    PointCloudEncoder,
    PoseEncoder,
)
from neuracore.ml.algorithm_utils.normalizer import MeanStdNormalizer

from .modules import ImageEncoder

PROPRIO_NORMALIZER = MeanStdNormalizer  # or MinMaxNormalizer
ACTION_NORMALIZER = MeanStdNormalizer  # or MinMaxNormalizer
RESNET_MEAN = [0.485, 0.456, 0.406]
RESNET_STD = [0.229, 0.224, 0.225]

LANGUAGE_MODEL_NAME = os.getenv("LANGUAGE_MODEL_NAME", "distilbert-base-uncased")
VISION_BACKBONE_DATA_TYPES = [
    DataType.RGB_IMAGES,
    DataType.DEPTH_IMAGES,
    DataType.POINT_CLOUDS,
]
PROPRIO_DATA_TYPES = [
    DataType.JOINT_POSITIONS,
    DataType.JOINT_VELOCITIES,
    DataType.JOINT_TORQUES,
    DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS,
]


class CNNMLP(NeuracoreModel):
    """CNN+MLP model with single timestep input and sequence output.

    A baseline model architecture that uses separate CNN encoders for each
    camera view, combines visual features with proprioceptive state, and
    predicts entire action sequences through a multi-layer perceptron.

    The model processes current observations and outputs a fixed-length
    sequence of future actions, making it suitable for action chunking
    approaches in robot manipulation.
    """

    def __init__(
        self,
        model_init_description: ModelInitDescription,
        image_backbone: str = "resnet18",
        use_resnet_stats: bool = True,
        hidden_dim: int = 512,
        cnn_output_dim: int = 512,
        num_layers: int = 3,
        lr: float = 1e-4,
        freeze_backbone: bool = False,
        lr_backbone: float = 1e-5,
        weight_decay: float = 1e-4,
    ):
        """Initialize the CNN+MLP model.

        Args:
            model_init_description: Model initialization parameters
            image_backbone: Backbone architecture for image encoders
            use_resnet_stats: Whether to use ResNet normalization statistics
            hidden_dim: Hidden dimension for MLP layers
            cnn_output_dim: Output dimension for CNN encoders
            num_layers: Number of MLP layers
            lr: Learning rate for main parameters
            freeze_backbone: Whether to freeze image encoder backbone
            lr_backbone: Learning rate for CNN backbone
            weight_decay: Weight decay for optimizer
        """
        super().__init__(model_init_description)
        self.image_backbone = image_backbone
        self.use_resnet_stats = use_resnet_stats
        self.hidden_dim = hidden_dim
        self.cnn_output_dim = cnn_output_dim
        self.num_layers = num_layers
        self.lr = lr
        self.freeze_backbone = freeze_backbone
        self.lr_backbone = lr_backbone
        self.weight_decay = weight_decay
        self.encoders = nn.ModuleDict()
        self.encoder_output_dims: dict[DataType, int] = {}

        data_stats: dict[DataType, DataItemStats] = {}

        if DataType.JOINT_POSITIONS in self.data_types:
            stats = self.dataset_statistics[DataType.JOINT_POSITIONS]
            stats = cast(list[JointDataStats], stats)
            combined_stats = DataItemStats()
            for stat in stats:
                combined_stats = combined_stats.concatenate(stat.value)
            data_stats[DataType.JOINT_POSITIONS] = combined_stats
            if DataType.JOINT_POSITIONS in self.input_data_types:
                self.encoder_output_dims[DataType.JOINT_POSITIONS] = cnn_output_dim
                self.encoders[DataType.JOINT_POSITIONS] = nn.Linear(
                    len(stats), cnn_output_dim
                )

        if DataType.JOINT_TARGET_POSITIONS in self.data_types:
            stats = self.dataset_statistics[DataType.JOINT_TARGET_POSITIONS]
            stats = cast(list[JointDataStats], stats)
            combined_stats = DataItemStats()
            for stat in stats:
                combined_stats = combined_stats.concatenate(stat.value)
            data_stats[DataType.JOINT_TARGET_POSITIONS] = combined_stats

        if DataType.JOINT_VELOCITIES in self.data_types:
            stats = self.dataset_statistics[DataType.JOINT_VELOCITIES]
            stats = cast(list[JointDataStats], stats)
            combined_stats = DataItemStats()
            for stat in stats:
                combined_stats = combined_stats.concatenate(stat.value)
            data_stats[DataType.JOINT_VELOCITIES] = combined_stats
            if DataType.JOINT_VELOCITIES in self.input_data_types:
                self.encoder_output_dims[DataType.JOINT_VELOCITIES] = cnn_output_dim
                self.encoders[DataType.JOINT_VELOCITIES] = nn.Linear(
                    len(stats), cnn_output_dim
                )

        if DataType.JOINT_TORQUES in self.data_types:
            stats = self.dataset_statistics[DataType.JOINT_TORQUES]
            stats = cast(list[JointDataStats], stats)
            combined_stats = DataItemStats()
            for stat in stats:
                combined_stats = combined_stats.concatenate(stat.value)
            data_stats[DataType.JOINT_TORQUES] = combined_stats
            if DataType.JOINT_TORQUES in self.input_data_types:
                self.encoder_output_dims[DataType.JOINT_TORQUES] = cnn_output_dim
                self.encoders[DataType.JOINT_TORQUES] = nn.Linear(
                    len(stats), cnn_output_dim
                )

        if DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS in self.data_types:
            stats = self.dataset_statistics[DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS]
            stats = cast(list[ParallelGripperOpenAmountDataStats], stats)
            combined_stats = DataItemStats()
            for stat in stats:
                combined_stats = combined_stats.concatenate(stat.open_amount)
            data_stats[DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS] = combined_stats
            if DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS in self.input_data_types:
                self.encoder_output_dims[DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS] = (
                    cnn_output_dim
                )
                self.encoders[DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS] = nn.Linear(
                    len(stats), cnn_output_dim
                )

        if DataType.RGB_IMAGES in self.input_data_types:
            stats = cast(
                list[CameraDataStats], self.dataset_statistics[DataType.RGB_IMAGES]
            )
            max_cameras = len(stats)
            self.encoder_output_dims[DataType.RGB_IMAGES] = max_cameras * cnn_output_dim
            self.encoders[DataType.RGB_IMAGES] = nn.ModuleList()
            for i in range(max_cameras):
                if use_resnet_stats:
                    mean, std = RESNET_MEAN, RESNET_STD
                else:
                    # Will be (3, H, W)
                    mean_c_h_w, std_c_h_w = stats[i].frame.mean, stats[i].frame.std
                    mean = mean_c_h_w.mean(axis=(1, 2)).tolist()
                    std = std_c_h_w.mean(axis=(1, 2)).tolist()
                encoder = torch.nn.Sequential(
                    T.Resize((224, 224)),
                    T.Normalize(mean=mean, std=std),
                    ImageEncoder(output_dim=cnn_output_dim, backbone=image_backbone),
                )
                self.encoders[DataType.RGB_IMAGES].append(encoder)

        if DataType.DEPTH_IMAGES in self.input_data_types:
            stats = cast(
                list[CameraDataStats], self.dataset_statistics[DataType.DEPTH_IMAGES]
            )
            max_cameras = len(stats)
            self.encoder_output_dims[DataType.DEPTH_IMAGES] = (
                max_cameras * cnn_output_dim
            )
            self.encoders[DataType.DEPTH_IMAGES] = nn.ModuleList()
            for i in range(max_cameras):
                encoder = torch.nn.Sequential(
                    T.Resize((224, 224)),
                    DepthImageEncoder(output_dim=cnn_output_dim),
                )
                self.encoders[DataType.DEPTH_IMAGES].append(encoder)

        if DataType.POINT_CLOUDS in self.input_data_types:
            stats = cast(
                list[PointCloudDataStats],
                self.dataset_statistics[DataType.POINT_CLOUDS],
            )
            max_pcs = len(stats)
            self.encoder_output_dims[DataType.POINT_CLOUDS] = max_pcs * cnn_output_dim
            self.encoders[DataType.POINT_CLOUDS] = nn.ModuleList()
            for i in range(max_pcs):
                encoder = PointCloudEncoder(output_dim=cnn_output_dim)
                self.encoders[DataType.POINT_CLOUDS].append(encoder)

        # All poses will share the same encoder
        if DataType.POSES in self.input_data_types:
            stats = self.dataset_statistics[DataType.POSES]
            stats = cast(list[DataItemStats], stats)
            max_poses = len(stats)
            self.encoder_output_dims[DataType.POSES] = cnn_output_dim
            self.encoders[DataType.POSES] = PoseEncoder(
                output_dim=cnn_output_dim,
                max_poses=max_poses,
            )

        # Language encoder (simplified - just use embedding)
        if DataType.LANGUAGE in self.input_data_types:
            from transformers import AutoTokenizer

            _tokenizer = AutoTokenizer.from_pretrained(LANGUAGE_MODEL_NAME)
            self.embedding_encoder = nn.Embedding(_tokenizer.vocab_size, 128)
            self.encoder_output_dims[DataType.LANGUAGE] = cnn_output_dim
            self.encoders[DataType.LANGUAGE] = nn.Sequential(
                nn.Linear(128, cnn_output_dim),
            )

        # Use multimodal fusion if multiple modalities
        self.fusion = MultimodalFusionEncoder(
            feature_dims=self.encoder_output_dims, output_dim=hidden_dim
        )
        mlp_input_dim = hidden_dim

        self.max_output_size = 0
        output_stats = []
        # Flatten norm_means into single parameters for outputs
        for data_type in self.output_data_types:
            output_stats.append(data_stats[data_type])
            self.max_output_size += len(self.dataset_statistics[data_type])

        input_stats = []
        self.proprio_dims = {}
        current_dim = 0
        for data_type in PROPRIO_DATA_TYPES:
            if data_type not in self.input_data_types:
                continue
            if data_type not in data_stats:
                continue
            input_stats.append(data_stats[data_type])
            dim = len(data_stats[data_type].mean)
            self.proprio_dims[data_type] = (current_dim, current_dim + dim)
            current_dim += dim

        self.action_normalizer = ACTION_NORMALIZER(
            name="actions", statistics=output_stats
        )
        self.proprio_normalizer = (
            PROPRIO_NORMALIZER(name="proprioception", statistics=input_stats)
            if input_stats
            else None
        )

        # Predict entire sequence at once
        self.output_size = self.max_output_size * self.output_prediction_horizon
        self.mlp = self._build_mlp(
            input_dim=mlp_input_dim,
            hidden_dim=hidden_dim,
            output_dim=self.output_size,
            num_layers=num_layers,
        )

        # Setup parameter groups
        self._setup_optimizer_param_groups()

    def _setup_optimizer_param_groups(self) -> None:
        """Setup parameter groups for optimizer."""
        backbone_params = []
        backbone_param_set = set()
        for data_type in VISION_BACKBONE_DATA_TYPES:
            if data_type in self.encoders:
                for param in self.encoders[data_type].parameters():
                    backbone_params.append(param)
                    backbone_param_set.add(param)
        other_params = [
            param for param in self.parameters() if param not in backbone_param_set
        ]

        if self.freeze_backbone:
            for param in backbone_params:
                param.requires_grad = False
            self.param_groups = [{"params": other_params, "lr": self.lr}]
        else:
            self.param_groups = [
                {"params": backbone_params, "lr": self.lr_backbone},
                {"params": other_params, "lr": self.lr},
            ]

    def _build_mlp(
        self, input_dim: int, hidden_dim: int, output_dim: int, num_layers: int
    ) -> nn.Sequential:
        """Construct multi-layer perceptron with normalization and dropout.

        Args:
            input_dim: Input feature dimension
            hidden_dim: Hidden layer dimension
            output_dim: Output dimension
            num_layers: Number of layers

        Returns:
            nn.Sequential: Constructed MLP module
        """
        if num_layers == 1:
            return nn.Sequential(nn.Linear(input_dim, output_dim))

        layers = [
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(),
            nn.LayerNorm(hidden_dim),
            nn.Dropout(0.1),
        ]

        for _ in range(num_layers - 2):
            layers.extend([
                nn.Linear(hidden_dim, hidden_dim),
                nn.ReLU(),
                nn.LayerNorm(hidden_dim),
                nn.Dropout(0.1),
            ])

        layers.append(nn.Linear(hidden_dim, output_dim))
        return nn.Sequential(*layers)

    def _encode_rgb_data(
        self,
        batched_nc_data: list[BatchedNCData],
        mask: torch.Tensor,
        encoders: nn.ModuleList,
    ) -> torch.Tensor:
        batched_rgb_data = cast(list[BatchedRGBData], batched_nc_data)
        assert len(batched_rgb_data) == len(
            encoders
        ), "Number of camera inputs does not match number of encoders."
        feats = []
        for encoder, input in zip(encoders, batched_rgb_data):
            last_frame = input.frame[:, -1, :, :, :]  # (B, 3, H, W)
            feats.append(encoder(last_frame))
        combined_feats = torch.stack(feats, dim=1)  # (B, num_cams, feat_dim)
        combined_feats *= mask.unsqueeze(-1)
        # (B, num_cams, feat_dim) -> (B, num_cams * feat_dim)
        return combined_feats.view(combined_feats.shape[0], -1)

    def _encode_depth_data(
        self,
        batched_nc_data: list[BatchedNCData],
        mask: torch.Tensor,
        encoders: nn.ModuleList,
    ) -> torch.Tensor:
        batched_depth_data = cast(list[BatchedDepthData], batched_nc_data)
        assert len(batched_depth_data) == len(
            encoders
        ), "Number of camera inputs does not match number of encoders."
        feats = []
        for encoder, input in zip(encoders, batched_depth_data):
            last_frame = input.frame[:, -1, :, :, :]  # (B, 1, H, W)
            feats.append(encoder(last_frame))
        combined_feats = torch.stack(feats, dim=1)  # (B, num_cams, feat_dim)
        combined_feats *= mask.unsqueeze(-1)
        # (B, num_cams, feat_dim) -> (B, num_cams * feat_dim)
        return combined_feats.view(combined_feats.shape[0], -1)

    def _encode_proprio(
        self,
        batch: BatchedInferenceInputs,
    ) -> dict[DataType, torch.Tensor]:
        """Encode all proprioceptive data with joint normalization."""
        # Concatenate all proprio data
        proprio_list = []
        for data_type in PROPRIO_DATA_TYPES:
            if data_type not in batch.inputs:
                continue

            batched_nc_data = batch.inputs[data_type]
            mask = batch.inputs_mask[data_type]

            if data_type == DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS:
                batched_gripper_data = cast(
                    list[BatchedParallelGripperOpenAmountData], batched_nc_data
                )
                proprio_data = torch.cat(
                    [bgd.open_amount for bgd in batched_gripper_data], dim=-1
                )
            else:
                batched_joint_data = cast(list[BatchedJointData], batched_nc_data)
                proprio_data = torch.cat(
                    [bjd.value for bjd in batched_joint_data], dim=-1
                )

            last_proprio = proprio_data[:, -1, :]  # (B, num_features)
            masked_proprio = last_proprio * mask
            proprio_list.append(masked_proprio)

        # Concatenate all proprio together: (B, total_proprio_dim)
        if not proprio_list:
            return {}

        all_proprio = torch.cat(proprio_list, dim=-1)

        # Normalize once on all proprio
        if self.proprio_normalizer is None:
            raise ValueError(
                "Proprioception inputs were provided but no statistics were available."
            )
        normalized_proprio = self.proprio_normalizer.normalize(all_proprio)

        # Split and encode each part
        features = {}
        for data_type, (start_idx, end_idx) in self.proprio_dims.items():
            if data_type in batch.inputs:
                proprio_slice = normalized_proprio[:, start_idx:end_idx]
                features[data_type] = self.encoders[data_type](proprio_slice)

        return features

    def _encode_pose_data(
        self,
        batched_nc_data: list[BatchedNCData],
        mask: torch.Tensor,
        encoder: nn.Module,
    ) -> torch.Tensor:
        batched_pose_data = cast(list[BatchedPoseData], batched_nc_data)
        # Gives (B, num_poses, pose_dim)
        pose_data = torch.stack([bpd.pose[:, -1] for bpd in batched_pose_data], dim=1)
        masked_pose_data = pose_data * mask.unsqueeze(-1)
        return encoder(masked_pose_data)  # (B, feat_dim)

    def _encode_point_cloud_data(
        self,
        batched_nc_data: list[BatchedNCData],
        mask: torch.Tensor,
        encoders: nn.ModuleList,
    ) -> torch.Tensor:
        batched_pc_data = cast(list[BatchedPointCloudData], batched_nc_data)
        assert len(batched_pc_data) == len(
            encoders
        ), "Number of point cloud inputs does not match number of encoders."
        feats = []
        for encoder, input in zip(encoders, batched_pc_data):
            last_points = input.points[:, -1, :, :]  # (B, 3, N)
            feats.append(encoder(last_points))
        combined_feats = torch.stack(feats, dim=1)  # (B, feat_dim, N)
        combined_feats *= mask.unsqueeze(-1)
        #  (B, num_points, feat_dim) -> (B, num_points * feat_dim)
        return combined_feats.view(combined_feats.shape[0], -1)

    def _encode_language_data(
        self,
        batched_nc_data: list[BatchedNCData],
        mask: torch.Tensor,
        encoder: nn.Module,
    ) -> torch.Tensor:
        """Encode language data using input_ids and attention_mask.

        Args:
            batched_nc_data: List of BatchedLanguageData instances
            mask: Mask tensor (not used for language data)
            encoder: Encoder module

        Returns:
            Encoded language features (B, final_dim)
        """
        batched_language_data = cast(list[BatchedLanguageData], batched_nc_data)

        # Grab the last language group and last timestep
        language_data = batched_language_data[-1]

        input_ids = language_data.input_ids  # (B, T, L)
        attention_mask = language_data.attention_mask  # (B, T, L)

        # Flatten B and T dimensions
        B, T, L = input_ids.shape
        input_ids = input_ids.view(B * T, L)  # (B*T, L)
        attention_mask = attention_mask.view(B * T, L)  # (B*T, L)

        # Embed tokens
        embeds = self.embedding_encoder(input_ids)  # (B*T, L, embed_dim)

        # Mean pooling with attention mask
        mask_expanded = attention_mask.unsqueeze(-1)  # (B*T, L, 1)
        sum_embeds = (embeds * mask_expanded).sum(dim=1)  # (B*T, embed_dim)
        sum_mask = mask_expanded.sum(dim=1).clamp(min=1e-9)  # (B*T, 1)
        pooled = sum_embeds / sum_mask  # (B*T, embed_dim)

        # Pass through MLP
        output = encoder[0](pooled)  # (B*T, output_dim)

        # Reshape back to (B, T, output_dim)
        output = output.view(B, T, -1)

        return output[:, -1, :]  # (B, final_dim)

    def _encode_inputs(
        self, batch: BatchedInferenceInputs
    ) -> dict[DataType, torch.Tensor]:
        """Process all visual modalities and return features."""
        features: dict[DataType, torch.Tensor] = {}

        # Encode proprio types together with joint normalization
        proprio_features = self._encode_proprio(batch)
        features.update(proprio_features)

        for data_type, batched_nc_data in batch.inputs.items():
            # Skip proprio types since they're already handled
            if data_type in PROPRIO_DATA_TYPES:
                continue

            mask = batch.inputs_mask[data_type]
            if data_type == DataType.RGB_IMAGES:
                features[data_type] = self._encode_rgb_data(
                    batched_nc_data,
                    mask,
                    cast(nn.ModuleList, self.encoders[data_type]),
                )
            elif data_type == DataType.DEPTH_IMAGES:
                features[data_type] = self._encode_depth_data(
                    batched_nc_data,
                    mask,
                    cast(nn.ModuleList, self.encoders[data_type]),
                )
            elif data_type == DataType.POSES:
                features[data_type] = self._encode_pose_data(
                    batched_nc_data,
                    mask,
                    self.encoders[data_type],
                )
            elif data_type == DataType.POINT_CLOUDS:
                features[data_type] = self._encode_point_cloud_data(
                    batched_nc_data,
                    mask,
                    cast(nn.ModuleList, self.encoders[data_type]),
                )
            elif data_type == DataType.LANGUAGE:
                features[data_type] = self._encode_language_data(
                    batched_nc_data,
                    mask,
                    self.encoders[data_type],
                )
        return features

    def _predict_action(self, batch: BatchedInferenceInputs) -> torch.FloatTensor:
        """Predict action sequence for the given batch.

        Processes visual and proprioceptive inputs through separate encoders,
        combines features, and predicts the entire action sequence through MLP.

        Args:
            batch: Input batch with observations

        Returns:
            torch.FloatTensor: Predicted action sequence [B, T, action_dim]
        """
        batch_size = len(batch)
        encoded_features = self._encode_inputs(batch)
        combined_features = self.fusion(encoded_features)

        # Forward through MLP to get entire sequence
        mlp_out = self.mlp(combined_features)
        action_preds = mlp_out.view(
            batch_size, self.output_prediction_horizon, self.max_output_size
        )
        return action_preds

    def forward(
        self, batch: BatchedInferenceInputs
    ) -> dict[DataType, list[BatchedNCData]]:
        """Perform inference to predict action sequence.

        Output will look like:
        {
            DataType.JOINT_TARGET_POSITIONS: [
                BatchedJointData,  # for each joint output
            ],
            DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS: [
                BatchedParallelGripperOpenAmountData,  # for each gripper output
            ],
            ...
        }

        Args:
            batch: Input batch with observations

        Returns:
            BatchedInferenceOutputs: Model predictions with action sequences
        """
        action_preds = self._predict_action(batch)
        # (B, T, action_dim)
        predictions = self.action_normalizer.unnormalize(action_preds)

        output_tensors: dict[DataType, list[BatchedNCData]] = {}
        start_slice_idx = 0
        for data_type in self.output_data_types:
            end_slice_idx = start_slice_idx + len(self.dataset_statistics[data_type])
            dt_preds = predictions[
                :, :, start_slice_idx:end_slice_idx
            ]  # (B, T, dt_size)
            if data_type in [DataType.JOINT_TARGET_POSITIONS, DataType.JOINT_POSITIONS]:
                batched_outputs = []
                for i in range(len(self.dataset_statistics[data_type])):
                    joint_preds = dt_preds[:, :, i : i + 1]  # (B, T, 1)
                    batched_outputs.append(BatchedJointData(value=joint_preds))
                output_tensors[data_type] = batched_outputs
            elif data_type == DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS:
                batched_outputs = []
                for i in range(len(self.dataset_statistics[data_type])):
                    gripper_preds = dt_preds[:, :, i : i + 1]  # (B, T, 1)
                    batched_outputs.append(
                        BatchedParallelGripperOpenAmountData(open_amount=gripper_preds)
                    )
                output_tensors[data_type] = batched_outputs
            else:
                raise ValueError(f"Unsupported output data type: {data_type}")
            start_slice_idx = end_slice_idx
        return output_tensors

    def training_step(self, batch: BatchedTrainingSamples) -> BatchedTrainingOutputs:
        """Perform a single training step.

        Predicts action sequences and computes mean squared error loss
        against target actions.

        Args:
            batch: Training batch with inputs and targets

        Returns:
            BatchedTrainingOutputs: Training outputs with losses and metrics
        """
        inference_sample = BatchedInferenceInputs(
            inputs=batch.inputs,
            inputs_mask=batch.inputs_mask,
            batch_size=batch.batch_size,
        )

        if set(batch.outputs.keys()) != set(self.output_data_types):
            raise ValueError(
                "Batch outputs do not match model output configuration."
                f" Expected {self.output_data_types}, got {list(batch.outputs.keys())}"
            )

        action_targets = []
        for i, data_type in enumerate(self.output_data_types):
            if data_type in [DataType.JOINT_TARGET_POSITIONS, DataType.JOINT_POSITIONS]:
                batched_joints = cast(list[BatchedJointData], batch.outputs[data_type])
                action_targets.extend([bjd.value for bjd in batched_joints])
            elif data_type == DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS:
                grippers = cast(
                    list[BatchedParallelGripperOpenAmountData], batch.outputs[data_type]
                )
                action_targets.extend([gripper.open_amount for gripper in grippers])
            else:
                raise ValueError(f"Unsupported output data type: {data_type}")

        action_data = torch.cat(action_targets, dim=-1)  # (B, 1, T * action_dim)
        action_data = action_data.view(
            batch.batch_size, self.output_prediction_horizon, -1
        )  # (B, T, action_dim)
        target_actions = self.action_normalizer.normalize(action_data)
        action_predictions = self._predict_action(inference_sample)

        losses: dict[str, Any] = {}
        metrics: dict[str, Any] = {}

        if self.training:
            losses["l1_loss"] = nn.functional.l1_loss(
                action_predictions, target_actions
            )

        return BatchedTrainingOutputs(
            losses=losses,
            metrics=metrics,
        )

    def configure_optimizers(
        self,
    ) -> list[torch.optim.Optimizer]:
        """Configure optimizer with different learning rates.

        Uses separate learning rates for image encoder backbone and other
        model parameters.

        Returns:
            list[torch.optim.Optimizer]: List of optimizers for model parameters
        """
        return [torch.optim.AdamW(self.param_groups, weight_decay=self.weight_decay)]

    @staticmethod
    def get_supported_input_data_types() -> set[DataType]:
        """Get the input data types supported by this model.

        Returns:
            set[DataType]: Set of supported input data types
        """
        return {
            DataType.JOINT_POSITIONS,
            DataType.JOINT_VELOCITIES,
            DataType.JOINT_TORQUES,
            DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS,
            DataType.POSES,
            DataType.RGB_IMAGES,
            DataType.DEPTH_IMAGES,
            DataType.POINT_CLOUDS,
            DataType.LANGUAGE,
        }

    @staticmethod
    def get_supported_output_data_types() -> set[DataType]:
        """Get the output data types supported by this model.

        Returns:
            set[DataType]: Set of supported output data types
        """
        return {
            DataType.JOINT_TARGET_POSITIONS,
            DataType.JOINT_POSITIONS,
            DataType.PARALLEL_GRIPPER_OPEN_AMOUNTS,
        }



"""Hydra-based training script for Neuracore models."""

import gc
import logging
import os
from pathlib import Path
from typing import Any

import hydra
import torch
import torch.multiprocessing as mp
from neuracore_types import ModelInitDescription
from neuracore_types.nc_data import DataType
from omegaconf import DictConfig, OmegaConf
from torch.utils.data import DataLoader, DistributedSampler, random_split

import neuracore as nc
from neuracore.api.training import _get_algorithms
from neuracore.core.utils.training_input_args_validation import (
    get_algorithm_name,
    validate_training_params,
)
from neuracore.ml import NeuracoreModel
from neuracore.ml.datasets.pytorch_single_sample_dataset import SingleSampleDataset
from neuracore.ml.datasets.pytorch_synchronized_dataset import (
    PytorchSynchronizedDataset,
)
from neuracore.ml.logging.cloud_training_logger import CloudTrainingLogger
from neuracore.ml.logging.tensorboard_training_logger import TensorboardTrainingLogger
from neuracore.ml.trainers.batch_autotuner import find_optimal_batch_size
from neuracore.ml.trainers.distributed_trainer import (
    DistributedTrainer,
    cleanup_distributed,
    setup_distributed,
)
from neuracore.ml.utils.algorithm_loader import AlgorithmLoader
from neuracore.ml.utils.algorithm_storage_handler import AlgorithmStorageHandler
from neuracore.ml.utils.device_utils import get_default_device
from neuracore.ml.utils.robot_data_spec_utils import (
    convert_str_to_robot_data_spec,
    extract_data_types,
    merge_robot_data_spec,
)
from neuracore.ml.utils.training_storage_handler import TrainingStorageHandler

# Environment setup
os.environ["PJRT_DEVICE"] = "GPU"

# Configure logging
logger = logging.getLogger(__name__)


def setup_logging(output_dir: str, rank: int = 0) -> None:
    """Setup logging configuration."""
    if rank == 0:
        output_path = Path(output_dir)
        output_path.mkdir(parents=True, exist_ok=True)
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s - %(levelname)s - %(message)s",
            handlers=[
                logging.StreamHandler(),
                logging.FileHandler(output_path / "train.log"),
            ],
        )
    else:
        # For other ranks, only log to console
        logging.basicConfig(
            level=logging.INFO,
            format=f"[Rank {rank}] %(asctime)s - %(levelname)s - %(message)s",
            handlers=[logging.StreamHandler()],
        )


def get_model_and_algorithm_config(
    cfg: DictConfig,
    model_init_description: ModelInitDescription,
) -> tuple[NeuracoreModel, dict[str, Any]]:
    """Get model and algorithm configuration."""
    algorithm_config: dict[str, Any] = {}
    if "algorithm" in cfg:
        algorithm_config = OmegaConf.to_container(cfg.algorithm, resolve=True)
        algorithm_config.pop("_target_", None)
        logger.info("Using custom algorithm parameters")
        logger.info(f"Algorithm parameters: {algorithm_config}")

        model = hydra.utils.instantiate(
            cfg.algorithm,
            model_init_description=model_init_description,
            **algorithm_config,
        )
    elif cfg.algorithm_id is not None:
        # Use algorithm_params for custom parameters
        if cfg.algorithm_params is not None:
            algorithm_config = OmegaConf.to_container(
                cfg.algorithm_params, resolve=True
            )
            logger.info("Using custom algorithm parameters")
            logger.info(f"Algorithm parameters: {algorithm_config}")

        extract_dir = Path(cfg.local_output_dir) / "algorithm"
        algorithm_loader = AlgorithmLoader(extract_dir)
        model_class = algorithm_loader.load_model()
        model = model_class(
            model_init_description=model_init_description,
            **algorithm_config,
        )
    else:
        raise ValueError(
            "Either 'algorithm' or 'algorithm_id' "
            "must be provided in the configuration"
        )
    return model, algorithm_config


def determine_optimal_batch_size(
    cfg: DictConfig,
    input_robot_data_spec: dict[str, dict[DataType, list[str]]],
    output_robot_data_spec: dict[str, dict[DataType, list[str]]],
    dataset: SingleSampleDataset,
    device: torch.device | None = None,
) -> int:
    """Run batch size autotuning on a single GPU and return the result."""
    if not torch.cuda.is_available() or (
        device is not None and "cuda" not in device.type
    ):
        raise ValueError("Autotuning is only supported on GPUs.")

    if device is None:
        device = get_default_device()

    logger.info(f"Starting batch size autotuning on {device}...")

    model_init_description = ModelInitDescription(
        dataset_statistics=dataset.dataset_statistics,
        input_data_types=extract_data_types(input_robot_data_spec),
        output_data_types=extract_data_types(output_robot_data_spec),
        output_prediction_horizon=cfg.output_prediction_horizon,
    )

    model, algorithm_config = get_model_and_algorithm_config(
        cfg, model_init_description
    )

    model = model.to(device)

    max_batch_size = cfg.max_batch_size if "max_batch_size" in cfg else len(dataset)
    min_batch_size = cfg.min_batch_size if "min_batch_size" in cfg else 2
    num_workers = cfg.batch_size_autotuning_num_workers

    logger.info(
        f"using max_batch_size: {max_batch_size}, "
        f"min_batch_size: {min_batch_size}, "
        f"num_workers: {num_workers}"
    )

    # Determine per-GPU batch size
    optimal_batch_size = find_optimal_batch_size(
        dataset=dataset,
        model=model,
        model_kwargs=algorithm_config,
        min_batch_size=min_batch_size,
        max_batch_size=max_batch_size,
        dataloader_kwargs={
            "collate_fn": dataset.collate_fn,
        },
    )

    # Clean up
    del model
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    gc.collect()

    logger.info(
        f"Autotuning complete. Optimal batch size per GPU: {optimal_batch_size}"
    )

    return optimal_batch_size


def run_training(
    rank: int,
    world_size: int,
    cfg: DictConfig,
    batch_size: int,
    input_robot_data_spec: dict[str, dict[DataType, list[str]]],
    output_robot_data_spec: dict[str, dict[DataType, list[str]]],
    dataset: PytorchSynchronizedDataset,
    device: torch.device | None = None,
) -> None:
    """Run the training process for a single GPU."""
    # Setup for distributed training
    if world_size > 1:
        nc.login()  # Ensure Neuracore is logged in on this process
        setup_distributed(rank, world_size)

    # Setup logging (different file per process)
    setup_logging(cfg.local_output_dir, rank)
    logger = logging.getLogger(__name__)

    # Set random seed (different for each process to ensure different data sampling)
    torch.manual_seed(cfg.seed + rank)

    try:
        logger.info(f"Using batch size: {batch_size}")

        # Merge data_types for synchronization
        merge_robot_data_spec(input_robot_data_spec, output_robot_data_spec)

        # Split dataset
        dataset_size = len(dataset)
        train_split = 1 - cfg.validation_split
        train_size = int(train_split * dataset_size)
        val_size = dataset_size - train_size

        # Use random split with fixed seed for deterministic behavior
        generator = torch.Generator().manual_seed(cfg.seed)
        train_dataset, val_dataset = random_split(
            dataset, [train_size, val_size], generator=generator
        )

        if world_size > 1:
            train_sampler = DistributedSampler(
                train_dataset,
                num_replicas=world_size,
                rank=rank,
                shuffle=True,
                seed=cfg.seed,
            )
            val_sampler = DistributedSampler(
                val_dataset,
                num_replicas=world_size,
                rank=rank,
                shuffle=False,
                seed=cfg.seed,
            )

            train_loader = DataLoader(
                train_dataset,
                batch_size=batch_size,
                sampler=train_sampler,
                num_workers=cfg.num_train_workers,
                pin_memory=True,
                persistent_workers=cfg.num_train_workers > 0,
                collate_fn=dataset.collate_fn,
            )

            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                sampler=val_sampler,
                num_workers=cfg.num_val_workers,
                pin_memory=True,
                persistent_workers=cfg.num_val_workers > 0,
                collate_fn=dataset.collate_fn,
            )
        else:
            # Regular data loaders for single GPU training
            train_loader = DataLoader(
                train_dataset,
                batch_size=batch_size,
                shuffle=True,
                num_workers=cfg.num_train_workers,
                pin_memory=True,
                persistent_workers=cfg.num_train_workers > 0,
                collate_fn=dataset.collate_fn,
            )

            val_loader = DataLoader(
                val_dataset,
                batch_size=batch_size,
                shuffle=False,
                num_workers=cfg.num_val_workers,
                pin_memory=True,
                persistent_workers=cfg.num_val_workers > 0,
                collate_fn=dataset.collate_fn,
            )

        # Log data loader information
        logger.info(
            f"Created data loaders with {len(train_loader.dataset)} training samples "
            f"and {len(val_loader.dataset)} validation samples"
        )

        # Model doesn't need to know about ids or names, just data types
        input_data_types = extract_data_types(input_robot_data_spec)
        output_data_types = extract_data_types(output_robot_data_spec)
        model_init_description = ModelInitDescription(
            dataset_statistics=dataset.dataset_statistics,
            input_data_types=input_data_types,
            output_data_types=output_data_types,
            output_prediction_horizon=cfg.output_prediction_horizon,
        )

        model, algorithm_config = get_model_and_algorithm_config(
            cfg, model_init_description
        )

        training_storage_handler = TrainingStorageHandler(
            local_dir=cfg.local_output_dir,
            training_job_id=cfg.training_id,
            algorithm_config=algorithm_config,
        )

        logger.info(
            f"Created model with "
            f"{sum(p.numel() for p in model.parameters()):,} parameters"
        )

        training_logger: TensorboardTrainingLogger | CloudTrainingLogger
        if cfg.training_id is None:
            training_logger = TensorboardTrainingLogger(
                log_dir=Path(cfg.local_output_dir) / "tensorboard",
            )
        else:
            training_logger = CloudTrainingLogger(
                training_id=cfg.training_id,
            )

        trainer = DistributedTrainer(
            model=model,
            train_loader=train_loader,
            val_loader=val_loader,
            training_logger=training_logger,
            storage_handler=training_storage_handler,
            output_dir=Path(cfg.local_output_dir),
            num_epochs=cfg.epochs,
            log_freq=cfg.logging_frequency,
            keep_last_n_checkpoints=cfg.keep_last_n_checkpoints,
            clip_grad_norm=algorithm_config.get("clip_grad_norm", None),
            rank=rank,
            world_size=world_size,
            device=device,
        )

        # Resume from checkpoint if specified
        start_epoch = 0
        if cfg.resume_checkpoint_path is not None:
            try:
                checkpoint = trainer.load_checkpoint(cfg.resume_checkpoint_path)
                start_epoch = checkpoint.get("epoch", 0) + 1
                logger.info(f"Resumed from checkpoint at epoch {start_epoch}")
            except Exception:
                logger.error("Failed to load checkpoint.", exc_info=True)

        # Start training
        try:
            logger.info("Starting training...")
            trainer.train(start_epoch=start_epoch)
            logger.info("Training completed successfully!")
        except Exception:
            logger.error("Training failed.", exc_info=True)
            raise

    finally:
        # Clean up distributed process group
        if world_size > 1:
            cleanup_distributed()

        logger.info(f"Process {rank} completed")


@hydra.main(version_base=None, config_path="config", config_name="config")
def main(cfg: DictConfig) -> None:
    """Main function to run the training script."""
    # Resolve the configuration
    OmegaConf.resolve(cfg)

    # Print configuration
    logger.info("Training configuration:")
    logger.info(OmegaConf.to_yaml(cfg, resolve=True))

    if "algorithm" in cfg and cfg.algorithm_id is not None:
        raise ValueError(
            "Both 'algorithm' and 'algorithm_id' are provided. "
            "Please specify only one."
        )
    if "algorithm" not in cfg and cfg.algorithm_id is None:
        raise ValueError(
            "Neither 'algorithm' nor 'algorithm_id' is provided. " "Please specify one."
        )

    if cfg.dataset_id is None and cfg.dataset_name is None:
        raise ValueError("Either 'dataset_id' or 'dataset_name' must be provided.")
    if cfg.dataset_id is not None and cfg.dataset_name is not None:
        raise ValueError(
            "Both 'dataset_id' and 'dataset_name' are provided. "
            "Please specify only one."
        )

    # Login and get dataset
    nc.login()
    if cfg.org_id is not None:
        nc.set_organization(cfg.org_id)

    if cfg.dataset_id is not None:
        dataset = nc.get_dataset(id=cfg.dataset_id)
    elif cfg.dataset_name is not None:
        dataset = nc.get_dataset(name=cfg.dataset_name)
    else:
        raise ValueError("Either 'dataset_id' or 'dataset_name' must be provided.")

    # Sort out data specs
    if cfg.input_robot_data_spec is not None:
        if not isinstance(cfg.input_robot_data_spec, DictConfig):
            raise ValueError(
                "'input_robot_data_spec' must be a dictionary "
                "mapping robot IDs to dictionary of data types to lists of data names."
            )
        input_robot_data_spec = convert_str_to_robot_data_spec(
            cfg.input_robot_data_spec
        )
    else:
        input_data_types = [DataType(dt) for dt in cfg.input_data_types]
        input_robot_data_spec = {
            robot_id: {dt: [] for dt in input_data_types}
            for robot_id in dataset.robot_ids
        }
    if cfg.output_robot_data_spec is not None:
        if not isinstance(cfg.output_robot_data_spec, DictConfig):
            raise ValueError(
                "'output_robot_data_spec' must either be None or a dictionary "
                "mapping robot IDs to dictions of data types to lists of data names."
            )
        output_robot_data_spec = convert_str_to_robot_data_spec(
            cfg.output_robot_data_spec
        )
    else:
        output_data_types = [DataType(dt) for dt in cfg.output_data_types]
        output_robot_data_spec = {
            robot_id: {dt: [] for dt in output_data_types}
            for robot_id in dataset.robot_ids
        }

    batch_size = cfg.batch_size

    algorithms_jsons = _get_algorithms()
    if cfg.algorithm_id is not None:
        algorithm_name = get_algorithm_name(
            algorithm_id=cfg.algorithm_id,
            algorithm_jsons=algorithms_jsons,
        )
    else:
        algorithm_name = cfg.algorithm._target_.rsplit(".", 1)[-1]

    validate_training_params(
        dataset,
        dataset_name=cfg.dataset_name if cfg.dataset_name is not None else "",
        algorithm_name=algorithm_name,
        input_robot_data_spec=input_robot_data_spec,
        output_robot_data_spec=output_robot_data_spec,
        algorithm_jsons=algorithms_jsons,
    )

    # Prepare data types for synchronization
    robot_data_spec = merge_robot_data_spec(
        input_robot_data_spec, output_robot_data_spec
    )

    synchronized_dataset = dataset.synchronize(
        frequency=cfg.frequency,
        robot_data_spec=robot_data_spec,
        prefetch_videos=True,
        max_prefetch_workers=cfg.max_prefetch_workers,
    )

    # Setup logging for main process
    setup_logging(cfg.local_output_dir)

    # Check if distributed training is enabled and multiple GPUs are available
    world_size = torch.cuda.device_count()

    if cfg.algorithm_id is not None:
        # Download the algorithm so that it can be processed later
        logger.info(f"Downloading algorithm from cloud with ID: {cfg.algorithm_id}")
        storage_handler = AlgorithmStorageHandler(algorithm_id=cfg.algorithm_id)
        extract_dir = Path(cfg.local_output_dir) / "algorithm"
        storage_handler.download_algorithm(extract_dir=extract_dir)
        logger.info(f"Algorithm extracted to {extract_dir}")

    device = None
    if cfg.device is not None:
        device = torch.device(cfg.device)
    else:
        device = get_default_device()

    # Create a pytorch synchronized dataset
    # NOTE: we are creating it here, and not in training to access the first sample
    # for batch size autotuning, if used.
    pytorch_dataset = PytorchSynchronizedDataset(
        synchronized_dataset=synchronized_dataset,
        input_robot_data_spec=input_robot_data_spec,
        output_robot_data_spec=output_robot_data_spec,
        output_prediction_horizon=cfg.output_prediction_horizon,
    )

    # Handle batch size configuration
    if isinstance(batch_size, str) and batch_size.lower() == "auto":
        sample = pytorch_dataset.load_sample(0).to(device)
        single_sample_dataset = SingleSampleDataset(
            sample=sample,
            input_robot_data_spec=input_robot_data_spec,
            output_robot_data_spec=output_robot_data_spec,
            output_prediction_horizon=cfg.output_prediction_horizon,
            dataset_statistics=pytorch_dataset.dataset_statistics,
            num_recordings=len(pytorch_dataset),
        )

        optimal_batch_size = determine_optimal_batch_size(
            cfg=cfg,
            input_robot_data_spec=input_robot_data_spec,
            output_robot_data_spec=output_robot_data_spec,
            dataset=single_sample_dataset,
            device=device,
        )
        batch_size = optimal_batch_size

    else:
        batch_size = int(batch_size)

    if world_size > 1:
        # Use multiprocessing to launch multiple processes
        mp.spawn(
            run_training,
            args=(
                world_size,
                cfg,
                batch_size,
                input_robot_data_spec,
                output_robot_data_spec,
                pytorch_dataset,
                device,
            ),
            nprocs=world_size,
            join=True,
        )
    else:
        # Single GPU or CPU training
        run_training(
            0,
            1,
            cfg,
            batch_size,
            input_robot_data_spec,
            output_robot_data_spec,
            pytorch_dataset,
            device,
        )


if __name__ == "__main__":
    main()



"""Interactive API key generation for Neuracore authentication.

This module provides functionality for generating API keys through interactive
user authentication. It prompts for email and password credentials, obtains
an access token, and creates a new API key for CLI usage.
"""

import argparse
from getpass import getpass

import requests
from pydantic import BaseModel, ValidationError

from neuracore.core.cli.get_user_confirmation import get_user_confirmation
from neuracore.core.config.config_manager import get_config_manager
from neuracore.core.exceptions import AuthenticationError, InputError

from ..const import API_URL, MAX_INPUT_ATTEMPTS


class APIKey(BaseModel):
    """API key model."""

    key: str


class Token(BaseModel):
    """OAuth2 token response model."""

    access_token: str
    token_type: str


def generate_api_key(email: str | None = None, password: str | None = None) -> str:
    """Generate a new API key through interactive user authentication.

    Prompts the user for their registered email and password, authenticates
    with the Neuracore server to obtain an access token, then uses that token
    to create a new API key for programmatic access. The process is interactive
    and handles authentication securely by hiding password input.

    Args:
        email: Optionally provide the email to try initially
        password: Optionally provide the password to try initially

    Returns:
        The generated API key string if successful, None if authentication
        or API key generation fails.

    Raises:
        AuthenticationError: If API key verification fails due to invalid
            credentials, network issues, or server errors.
        InputError: If there is an issue with the user's input
    """
    # Prompt the user for credentials
    access_token = None

    for i in range(MAX_INPUT_ATTEMPTS):
        try:
            if not email:
                email = input("Enter your registered email: ")
            if not password:
                password = getpass("Enter your password: ")

            auth_response = requests.post(
                f"{API_URL}/auth/token",
                headers={"Content-Type": "application/x-www-form-urlencoded"},
                data={"username": email, "password": password},
            )

            # Handle 401 *before* raise_for_status so it doesn't become HTTPError
            if auth_response.status_code == 401:
                # If this was the last allowed attempt, raise "too many attempts"
                if i + 1 >= MAX_INPUT_ATTEMPTS:
                    print("Too many failed attempts, perhaps check your Caps Lock?")
                    raise InputError("Out of attempts")

                print("Incorrect email or password.")
                again = get_user_confirmation("Do you wish to try again?")
                if again:
                    email = None
                    password = None
                    continue
                raise InputError("Invalid Email or Password")

            auth_response.raise_for_status()
            token_data = Token.model_validate(auth_response.json())
            access_token = token_data.access_token
            break

        except KeyboardInterrupt:
            raise InputError("User cancelled the operation.")
        except ValidationError:
            raise AuthenticationError("Invalid token from server")
        except requests.exceptions.ConnectionError:
            raise AuthenticationError(
                "Failed to connect to neuracore server, "
                "please check your internet connection and try again."
            )
        except requests.exceptions.RequestException as e:
            raise AuthenticationError(f"Failed to get Auth Token: {e}")

    # Use the access token to request an API key
    try:
        headers = {"Authorization": f"Bearer {access_token}"}
        api_key_response = requests.get(
            f"{API_URL}/auth/api-key",
            headers=headers,
        )
        api_key_response.raise_for_status()
        api_key = APIKey.model_validate(api_key_response.json()).key
        print(f"Your API key is: {api_key}")
        config_manager = get_config_manager()
        config_manager.config.api_key = api_key
        config_manager.save_config()
        return api_key
    except ValidationError:
        raise AuthenticationError("Invalid api-key from server")
    except requests.exceptions.ConnectionError:
        raise AuthenticationError(
            "Failed to connect to neuracore server, "
            "please check your internet connection and try again."
        )
    except requests.exceptions.RequestException as e:
        raise AuthenticationError(f"Failed to create API key: {e}")


def main() -> None:
    """Main function to run the API key generation process."""
    try:
        parser = argparse.ArgumentParser(
            description="Generate a new API key for access to the neuracore platform.",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        parser.add_argument(
            "--email",
            "--username",
            "-e",
            dest="email",
            required=False,
            type=str,
            help="The email to login with.",
        )
        parser.add_argument(
            "--password",
            "-p",
            dest="password",
            required=False,
            type=str,
            help="The password to login with.",
        )
        args = parser.parse_args()
        generate_api_key(email=args.email, password=args.password)
    except AuthenticationError:
        print("Failed to generate API key, please try again later")
    except InputError:
        print("Failed to generate API key due to incorrect input")


if __name__ == "__main__":
    main()



"""Resumable upload functionality for Google Cloud Storage integration.

This module provides a ResumableUpload class that handles chunked uploads to
Google Cloud Storage with retry logic and progress tracking. It supports
resuming interrupted uploads and provides proper error handling with
exponential backoff.
"""

import logging
import time

import requests

from neuracore.core.auth import get_auth
from neuracore.core.config.get_current_org import get_current_org
from neuracore.core.const import API_URL

logger = logging.getLogger(__name__)


class ResumableUpload:
    """Handles resumable uploads to Google Cloud Storage.

    This class manages the lifecycle of a resumable upload session, including
    obtaining upload URLs, tracking progress, handling retries, and managing
    chunk uploads with proper HTTP range headers. It supports resuming
    interrupted uploads by checking server status.
    """

    def __init__(self, recording_id: str, filepath: str, content_type: str):
        """Initialize a resumable upload to GCS.

        Args:
            recording_id: Unique identifier for the recording.
            filepath: Target file path in the storage bucket.
            content_type: MIME type of the file being uploaded.
        """
        self.recording_id = recording_id
        self.filepath = filepath
        self.content_type = content_type
        self.session_uri = self._get_upload_session_uri()
        self.total_bytes_uploaded = 0
        self.max_retries = 5

    def _get_upload_session_uri(self) -> str:
        """Get a resumable upload session URI from the backend.

        Makes an API call to obtain a resumable upload session URL that will
        be used for all subsequent chunk uploads.

        Returns:
            The resumable upload session URI from Google Cloud Storage.

        Raises:
            requests.HTTPError: If the API request fails.
        """
        auth = get_auth()
        params = {
            "filepath": self.filepath,
            "content_type": self.content_type,
        }
        org_id = get_current_org()
        response = requests.get(
            f"{API_URL}/org/{org_id}/recording/{self.recording_id}/resumable_upload_url",
            params=params,
            headers=auth.get_headers(),
        )
        response.raise_for_status()
        return response.json()["url"]

    def upload_chunk(self, data: bytes, is_final: bool = False) -> bool:
        """Upload a chunk of data to the resumable upload session.

        Uploads a chunk of data with proper range headers and retry logic.
        Automatically handles upload position verification and exponential
        backoff on failures.

        Args:
            data: Binary data chunk to upload.
            is_final: Whether this is the final chunk of the upload.

        Returns:
            True if the upload was successful, False otherwise.

        Raises:
            Exception: If there's a mismatch between local and server upload positions.
        """
        if len(data) == 0 and not is_final:
            return True  # Nothing to upload

        # First, check if the session is still valid and get current uploaded bytes
        actual_uploaded_bytes = self.check_status()
        if actual_uploaded_bytes != self.total_bytes_uploaded:
            raise Exception(
                "Upload position mismatch: "
                f"Local={self.total_bytes_uploaded}, Server={actual_uploaded_bytes}"
            )

        # Prepare headers
        headers = {
            "Content-Length": str(len(data)),
        }

        # Set content range header
        chunk_first_byte = self.total_bytes_uploaded
        chunk_last_byte = self.total_bytes_uploaded + len(data) - 1

        if is_final:
            # Final chunk, include total size
            total_size = self.total_bytes_uploaded + len(data)
            headers["Content-Range"] = (
                f"bytes {chunk_first_byte}-{chunk_last_byte}/{total_size}"
            )
        else:
            # Not final chunk, use '*' for total size
            headers["Content-Range"] = f"bytes {chunk_first_byte}-{chunk_last_byte}/*"

        # Attempt the upload with retries
        for attempt in range(self.max_retries):
            try:
                response = requests.put(self.session_uri, headers=headers, data=data)
                status_code = response.status_code

                if status_code == 200 or status_code == 201:
                    self.total_bytes_uploaded += len(data)
                    return True
                elif status_code == 308:
                    # Resume Incomplete, more data expected
                    self.total_bytes_uploaded += len(data)
                    return True
                else:
                    # Error occurred
                    if attempt < self.max_retries - 1:
                        time.sleep(2**attempt)  # Exponential backoff

            except Exception:
                logger.error(
                    f"Exception during upload (attempt {attempt+1})", exc_info=True
                )
                if attempt < self.max_retries - 1:
                    time.sleep(2**attempt)  # Exponential backoff

        return False

    def check_status(self) -> int:
        """Check the status of the resumable upload session.

        Queries the upload session to determine how many bytes have been
        successfully uploaded so far. This is used for resuming interrupted
        uploads and verifying upload progress.

        Returns:
            Number of bytes that have been successfully uploaded to the server.

        Raises:
            Exception: If the server returns an unexpected status code.
        """
        headers = {"Content-Length": "0", "Content-Range": "bytes */*"}

        response = requests.put(self.session_uri, headers=headers)
        if response.status_code == 200 or response.status_code == 201:
            logger.debug("Upload complete")
            return self.total_bytes_uploaded
        elif response.status_code == 308 and "Range" in response.headers:
            range_header = response.headers["Range"]
            return int(range_header.split("-")[1]) + 1
        elif response.status_code == 308:
            return 0

        raise Exception(f"Unexpected status code: {response.status_code}")



"""Lightweight FastAPI server for local model inference.

This replaces TorchServe with a more flexible, custom solution that gives us
full control over the inference pipeline while maintaining .nc.zip compatibility.
"""

import json
import logging
import time
from pathlib import Path

import uvicorn
from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from neuracore_types import BatchedNCDataUnion, DataType, SynchronizedPoint
from pydantic import BaseModel

from neuracore.core.exceptions import InsufficientSynchronizedPointError

logger = logging.getLogger(__name__)

PING_ENDPOINT = "/ping"
PREDICT_ENDPOINT = "/predict"
SET_CHECKPOINT_ENDPOINT = "/set_checkpoint"


class CheckpointRequest(BaseModel):
    """Request model for setting checkpoints."""

    epoch: int


class ModelServer:
    """Lightweight model server using FastAPI."""

    def __init__(
        self,
        model_input_order: dict[DataType, list[str]],
        model_output_order: dict[DataType, list[str]],
        model_file: Path,
        org_id: str,
        job_id: str | None = None,
        device: str | None = None,
    ):
        """Initialize the model server.

        Args:
            model_input_order: Model input order
            model_output_order: Model output order
            model_file: Path to the .nc.zip model archive
            org_id: Organization ID for the model
            job_id: Job ID for the model
            device: Device the model loaded on
        """
        # Import here to avoid the need for pytorch unless the user uses this policy
        from neuracore.ml.utils.policy_inference import PolicyInference

        self.policy_inference = PolicyInference(
            model_input_order=model_input_order,
            model_output_order=model_output_order,
            org_id=org_id,
            job_id=job_id,
            model_file=model_file,
            device=device,
        )
        self.app = self._create_app()

    def _create_app(self) -> FastAPI:
        """Create and configure the FastAPI application."""
        app = FastAPI(
            title="Neuracore Model Server",
            description="Lightweight model inference server",
            version="1.0.0",
        )

        # Add CORS middleware
        app.add_middleware(
            CORSMiddleware,
            allow_origins=["*"],
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )

        # Health check endpoint
        @app.get(PING_ENDPOINT)
        async def health_check() -> dict:
            return {"status": "healthy", "timestamp": time.time()}

        # Main prediction endpoint
        @app.post(
            PREDICT_ENDPOINT,
            response_model=dict[DataType, dict[str, BatchedNCDataUnion]],
        )
        async def predict(
            sync_point: SynchronizedPoint,
        ) -> dict[DataType, dict[str, BatchedNCDataUnion]]:
            try:
                return self.policy_inference(sync_point)
            except InsufficientSynchronizedPointError:
                logger.error("Insufficient sync point data.")
                raise HTTPException(
                    status_code=422,
                    detail="Insufficient sync point data for inference.",
                )
            except Exception as e:
                logger.error("Prediction error.", exc_info=True)
                raise HTTPException(
                    status_code=500, detail=f"Prediction failed: {str(e)}"
                )

        @app.post(SET_CHECKPOINT_ENDPOINT)
        async def set_checkpoint(request: CheckpointRequest) -> None:
            try:
                self.policy_inference.set_checkpoint(request.epoch)
            except Exception as e:
                logger.error("Checkpoint loading error.", exc_info=True)
                raise HTTPException(
                    status_code=500, detail=f"Checkpoint loading failed: {str(e)}"
                )

        return app

    def run(
        self, host: str = "0.0.0.0", port: int = 8080, log_level: str = "info"
    ) -> None:
        """Run the server.

        Args:
            host: Host to bind to
            port: Port to bind to
            log_level: Logging level
        """
        uvicorn.run(
            self.app, host=host, port=port, log_level=log_level, access_log=True
        )


def start_server(
    model_input_order: dict[DataType, list[str]],
    model_output_order: dict[DataType, list[str]],
    model_file: Path,
    org_id: str,
    job_id: str | None = None,
    host: str = "0.0.0.0",
    port: int = 8080,
    log_level: str = "info",
    device: str | None = None,
) -> ModelServer:
    """Start a model server instance.

    Args:
        model_file: Path to the .nc.zip model archive
        org_id: Organization ID
        job_id: Job ID
        host: Host to bind to
        port: Port to bind to
        log_level: Logging level
        device: Device model loaded on

    Returns:
        ModelServer instance
    """
    server = ModelServer(
        model_input_order=model_input_order,
        model_output_order=model_output_order,
        model_file=model_file,
        org_id=org_id,
        job_id=job_id,
        device=device,
    )
    server.run(host, port, log_level)
    return server


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(description="Start Neuracore Model Server")
    parser.add_argument(
        "--model-input-order",
        required=True,
        help=(
            "Model input order consisting of json dump of "
            "dict mapping DataType to list of strings"
        ),
    )
    parser.add_argument(
        "--model-output-order",
        required=True,
        help=(
            "Model output order consisting of json dump of "
            "dict mapping DataType to list of strings"
        ),
    )
    parser.add_argument(
        "--model-file", required=True, help="Path to .nc.zip model file"
    )
    parser.add_argument("--org-id", required=True, help="Organization ID")
    parser.add_argument("--job-id", required=False, help="Job ID")
    parser.add_argument("--host", default="0.0.0.0", help="Host to bind to")
    parser.add_argument("--port", type=int, default=8080, help="Port to bind to")
    parser.add_argument("--log-level", default="info", help="Logging level")
    parser.add_argument("--device", help="Device to load model on (cpu, cuda, etc.)")

    args = parser.parse_args()

    model_input_order = {
        DataType(k): v for k, v in json.loads(str(args.model_input_order)).items()
    }
    model_output_order = {
        DataType(k): v for k, v in json.loads(str(args.model_output_order)).items()
    }

    start_server(
        model_input_order=model_input_order,
        model_output_order=model_output_order,
        model_file=Path(args.model_file),
        org_id=args.org_id,
        job_id=args.job_id,
        host=args.host,
        port=args.port,
        log_level=args.log_level,
        device=args.device,
    )



"""Image encoder module using pretrained ResNet backbone.

This module provides a simple image encoder that uses pretrained ResNet
architectures for feature extraction, followed by a projection layer to
map features to a desired output dimension.
"""

import torch
import torch.nn as nn
import torchvision.models as models


class ImageEncoder(nn.Module):
    """Encode images using ResNet backbone with projection layer.

    Uses a pretrained ResNet architecture (without the final classification
    layer) to extract visual features, followed by a linear projection to
    map to the desired output dimension. Suitable for various computer
    vision tasks requiring fixed-size feature representations.
    """

    OUTPUT_CNN_W_H = 7 * 7

    def __init__(self, output_dim: int = 512, backbone: str = "resnet18"):
        """Initialize the image encoder.

        Args:
            output_dim: Desired output feature dimension
            backbone: ResNet architecture name (e.g., "resnet18", "resnet50")
        """
        super().__init__()
        # Use pretrained ResNet but remove final layer
        self.backbone, fc_input_dim = self._build_backbone(backbone)
        self.proj_2d = nn.Conv2d(fc_input_dim, output_dim, kernel_size=1)
        self.linear = nn.Linear(ImageEncoder.OUTPUT_CNN_W_H * output_dim, output_dim)

    def _build_backbone(self, backbone_name: str) -> tuple[nn.Module, int]:
        """Build backbone CNN by removing classification layers.

        Args:
            backbone_name: Name of the ResNet architecture to use

        Returns:
            nn.Module: ResNet backbone without final classification layers
            int: Input dimension of the final fully connected layer
        """
        resnet = models.get_model(backbone_name, weights="DEFAULT")
        return nn.Sequential(*list(resnet.children())[:-2]), resnet.fc.in_features

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass through image encoder.

        Processes input images through the ResNet backbone, flattens the
        spatial dimensions, and projects to the desired output dimension.

        Args:
            x: Image tensor of shape (batch, channels, height, width)

        Returns:
            torch.Tensor: Encoded features of shape (batch, output_dim)
        """
        batch = x.shape[0]
        x = self.backbone(x)
        x = self.proj_2d(x).view(batch, -1)
        return self.linear(x)



"""Synchronized recording iterator."""

import logging
import shutil
import subprocess
import sys
import tempfile
import time
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, cast

import numpy as np
import requests
import wget
from neuracore_types import CameraData, DataType, RobotDataSpec, SynchronizationDetails
from neuracore_types import SynchronizedEpisode as SynchronizedEpisodeModel
from neuracore_types import SynchronizedPoint, SynchronizeRecordingRequest
from PIL import Image

from neuracore.core.data.cache_manager import CacheManager

from ..auth import get_auth
from ..const import API_URL

logger = logging.getLogger(__name__)

if TYPE_CHECKING:
    from neuracore.core.data.dataset import Dataset

MAX_DECODING_ATTEMPTS = 3
_FFMPEG_AVAILABLE: bool | None = None


class SynchronizedRecording:
    """Synchronized recording iterator."""

    def __init__(
        self,
        dataset: "Dataset",
        recording_id: str,
        robot_id: str,
        instance: int,
        frequency: int = 0,
        robot_data_spec: RobotDataSpec | None = None,
        prefetch_videos: bool = False,
    ):
        """Initialize episode iterator for a specific recording.

        Args:
            dataset: Parent Dataset instance.
            recording_id: Recording ID string.
            robot_id: The robot that created this recording.
            instance: The instance of the robot that created this recording.
            frequency: Frequency at which to synchronize the recording.
            robot_data_spec: Robot data specification for synchronization.
            prefetch_videos: Whether to prefetch video data to cache on initialization.
        """
        self.dataset = dataset
        self.id = recording_id
        self.frequency = frequency
        self.robot_data_spec = robot_data_spec
        self.cache_dir: Path = dataset.cache_dir
        self.robot_id = robot_id
        self.instance = instance

        self._episode_synced = self._get_synced_data()
        self._episode_length = len(self._episode_synced.observations)

        # Use start_time and end_time from the synchronized episode,
        # as they reflect trim_start_end settings from synchronization
        self.start_time = self._episode_synced.start_time
        self.end_time = self._episode_synced.end_time
        self.cache_manager = CacheManager(
            self.cache_dir,
        )
        self._iter_idx = 0
        self._suppress_wget_progress = True

        if prefetch_videos:
            cache = self.dataset.cache_dir / self.id
            # Check if cache directory exists and contains any files
            self._wait_for_lock_release(cache / ".recording.lock", cache)
            # NOTE: this is to start video prefetching frames into cache
            self._get_sync_point(0)

    def _get_synced_data(self) -> SynchronizedEpisodeModel:
        """Retrieve synchronized metadata for the recording.

        Returns:
            SynchronizedEpisode object containing synchronized frames and metadata.

        Raises:
            requests.HTTPError: If the API request fails.
        """
        auth = get_auth()
        response = requests.post(
            f"{API_URL}/org/{self.dataset.org_id}/synchronize/synchronize-recording",
            json=SynchronizeRecordingRequest(
                recording_id=self.id,
                synchronization_details=SynchronizationDetails(
                    frequency=self.frequency,
                    robot_data_spec=self.robot_data_spec,
                    max_delay_s=sys.float_info.max,
                    allow_duplicates=True,
                ),
            ).model_dump(mode="json"),
            headers=auth.get_headers(),
        )
        response.raise_for_status()
        return SynchronizedEpisodeModel.model_validate(response.json())

    def _get_video_url(self, camera_type: DataType, camera_id: str) -> str:
        """Get streaming URL for a specific camera's video data.

        Args:
            camera_type: Type of camera (e.g., "rgbs", "depths").
            camera_id: Unique identifier for the camera.

        Returns:
            URL string for downloading the video file.

        Raises:
            requests.HTTPError: If the API request fails.
        """
        auth = get_auth()
        response = requests.get(
            f"{API_URL}/org/{self.dataset.org_id}/recording/{self.id}/download_url",
            params={"filepath": f"{camera_type.value}/{camera_id}/lossless.mp4"},
            headers=auth.get_headers(),
        )
        response.raise_for_status()
        return response.json()["url"]

    def _decode_video(self, video_location: Path, video_frame_cache_path: Path) -> None:
        """Extract frames from video and cache them to disk.

        Args:
            video_location: Path to the video file.
            video_frame_cache_path: Path to the directory where video frames are cached.
        """
        """Extract frames from video and cache them to disk."""
        global _FFMPEG_AVAILABLE

        # Lazily determine ffmpeg availability once
        if _FFMPEG_AVAILABLE is None:
            try:
                subprocess.run(
                    ["ffmpeg", "-version"],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    check=True,
                )
                _FFMPEG_AVAILABLE = True
            except (FileNotFoundError, subprocess.CalledProcessError):
                _FFMPEG_AVAILABLE = False
                logger.warning(
                    "ffmpeg not found. Falling back to PyAV for video decoding. "
                    "Install ffmpeg for significantly faster decoding."
                )

        if _FFMPEG_AVAILABLE:
            output_pattern = str(video_frame_cache_path / "%d.png")
            try:
                subprocess.run(
                    [
                        "ffmpeg",
                        "-i",
                        str(video_location),
                        "-vsync",
                        "0",
                        "-q:v",
                        "1",
                        "-start_number",
                        "0",
                        output_pattern,
                        "-y",
                        "-loglevel",
                        "error",
                    ],
                    check=True,
                    capture_output=True,
                )
                return
            except subprocess.CalledProcessError:
                logger.error("ffmpeg failed during decoding, falling back to PyAV")
                _FFMPEG_AVAILABLE = False  # Permanently disable ffmpeg for this run

        # PyAV fallback (executed only once ffmpeg is known unavailable)
        import av

        with av.open(str(video_location)) as container:
            for i, frame in enumerate(container.decode(video=0)):
                frame_image = Image.fromarray(frame.to_rgb().to_ndarray())
                frame_file = video_frame_cache_path / f"{i}.png"
                frame_image.save(frame_file)

    def _download_video_and_cache_frames_to_disk(
        self, camera_type: DataType, camera_id: str, video_frame_cache_path: Path
    ) -> None:
        """Download video and cache individual frames as images.

        Args:
            camera_type: Type of camera (e.g., "rgbs", "depths").
            camera_id: Unique identifier for the camera.
            video_frame_cache_path: Path to the directory where video frames are cached.
        """
        lock_file = video_frame_cache_path / ".recording.lock"
        lock_acquired = self._create_decoding_lock(lock_file, camera_id)

        try:
            # Create a temporary video file path
            self.cache_manager.ensure_space_available()

            with tempfile.TemporaryDirectory() as temp_dir:
                # Download video to temporary directory
                video_location = Path(temp_dir) / f"{camera_id}{camera_type.value}.mp4"
                wget.download(
                    self._get_video_url(camera_type, camera_id),
                    str(video_location),
                    bar=None if self._suppress_wget_progress else wget.bar_thermometer,
                )
                # Decode video to frames and cache them to disk
                self._decode_video(video_location, video_frame_cache_path)
        finally:
            if lock_acquired:
                self._delete_decoding_lock(lock_file)

    def _create_decoding_lock(self, lock_file: Path, camera_id: str) -> bool:
        """Create an exclusive lock file for decoding."""
        try:
            # Create the lock file exclusively
            lock_file.parent.mkdir(parents=True, exist_ok=True)
            lock_file.touch(exist_ok=False)
        except FileExistsError as exc:
            raise RuntimeError(
                f"Another process is already decoding video for camera {camera_id}"
            ) from exc
        return True

    def _delete_decoding_lock(self, lock_file: Path) -> None:
        """Remove the decoding lock file if present."""
        lock_file.unlink(missing_ok=True)

    def _check_stale_lock_file(self, lock_file: Path, timeout: int = 300) -> bool:
        """Check if a lock file is stale based on a timeout.

        Args:
            lock_file: Path to the lock file.
            timeout: Time in seconds after which the lock is considered stale.
                    (default: 300s/5min)

        Returns:
            True if the lock file is stale, False otherwise.
        """
        if not lock_file.exists():
            return False
        lock_mtime = lock_file.stat().st_mtime
        if (time.time() - lock_mtime) > timeout:
            return True
        return False

    def _wait_for_lock_release(
        self, lock_file: Path, parent_folder_path: Path, check_interval: int = 1
    ) -> None:
        """Wait for a lock file to be released.

        Args:
            lock_file: Path to the lock file.
            parent_folder_path: Path to the parent folder containing the lock file.
            check_interval: Time in seconds between checks.
        """
        # Check if the lock is stale
        while lock_file.exists():
            if self._check_stale_lock_file(lock_file):
                logger.warning(
                    f"Stale lock file detected at {lock_file}. Removing lock."
                )
                self._delete_decoding_lock(lock_file)
                shutil.rmtree(parent_folder_path, ignore_errors=True)
                logger.info(
                    f"Removed stale lock and cleared cache at {parent_folder_path}."
                )
                break
            time.sleep(check_interval)

    def _get_frame_from_disk_cache(
        self,
        camera_type: DataType,
        camera_data: dict[str, CameraData],
        transform_fn: Callable[[np.ndarray], np.ndarray] | None = None,
    ) -> dict[str, CameraData]:
        """Get video frame from disk cache for camera data.

        Args:
            camera_type: DataType indicating the type of camera data.
            camera_data: Dictionary of camera data with camera IDs as keys.
            frame_idx: Index of the frame to retrieve.
            transform_fn: Optional function to transform frames (e.g., rgb_to_depth).

        Returns:
            Dictionary of CameraData with populated frames.
        """
        # Create new dict with new CameraData instances to avoid mutating originals
        result = {}
        for cam_id, cam_data in camera_data.items():
            cam_id_rgb_root = self.cache_dir / f"{self.id}" / camera_type.value / cam_id
            lock_file = cam_id_rgb_root / ".recording.lock"
            self._wait_for_lock_release(lock_file, cam_id_rgb_root)

            if not cam_id_rgb_root.exists():
                # Not in cache, download video and cache frames to disk
                cam_id_rgb_root.mkdir(parents=True, exist_ok=True)
                self._download_video_and_cache_frames_to_disk(
                    camera_type, cam_id, cam_id_rgb_root
                )

            frame_file = cam_id_rgb_root / f"{cam_data.frame_idx}.png"
            frame = Image.open(frame_file)

            if transform_fn:
                frame = Image.fromarray(transform_fn(np.array(frame)))

            result[cam_id] = cam_data.model_copy(update={"frame": frame})

        return result

    def _insert_camera_data_intro_sync_point(
        self, sync_point: SynchronizedPoint
    ) -> SynchronizedPoint:
        """Populate video frames for a given sync point.

        Args:
            sync_point: SynchronizedPoint object containing
                camera data (without frames).

        Returns:
            A new SynchronizedPoint object with populated video frames.
        """
        # Build new data dict with loaded frames
        new_data = {}
        for data_type, data_dict in sync_point.data.items():
            if data_type == DataType.RGB_IMAGES:
                new_data[data_type] = self._get_frame_from_disk_cache(
                    DataType.RGB_IMAGES, data_dict
                )
            elif data_type == DataType.DEPTH_IMAGES:
                new_data[data_type] = self._get_frame_from_disk_cache(
                    DataType.DEPTH_IMAGES, data_dict
                )
            else:
                # create NEW instances to avoid shared references
                new_data[data_type] = {
                    name: nc_data.model_copy() for name, nc_data in data_dict.items()
                }

        return SynchronizedPoint(
            timestamp=sync_point.timestamp,
            robot_id=sync_point.robot_id,
            data=new_data,
        )

    def _get_sync_point(self, idx: int) -> SynchronizedPoint:
        """Get synchronized data point at a specific index.

        Args:
            idx: Index of the sync point to retrieve.

        Returns:
            SynchronizedPoint object containing synchronized data
                for the specified index.
        """
        sync_point = self._episode_synced.observations[idx]
        sync_point = self._insert_camera_data_intro_sync_point(sync_point)
        return sync_point

    def __iter__(self) -> "SynchronizedRecording":
        """Initialize iteration over the episode.

        Returns:
            SynchronizedRecording instance for iteration.
        """
        self._iter_idx = 0
        return self

    def __len__(self) -> int:
        """Get the number of timesteps in the episode.

        Returns:
            int: Number of timesteps in the episode.
        """
        return self._episode_length

    def __getitem__(
        self, idx: int | slice
    ) -> SynchronizedPoint | list[SynchronizedPoint]:
        """Support for indexing episode data.

        Args:
            idx: Integer index or slice object for accessing sync points.

        Returns:
            SynchronizedPoint object for single index or list of
                SynchronizedPoint objects for slice.

        Raises:
            IndexError: If the index is out of range.
            TypeError: If the index is not an integer or slice.
        """
        if isinstance(idx, slice):
            # Handle slice objects
            start, stop, step = idx.indices(len(self))
            return [cast(SynchronizedPoint, self[i]) for i in range(start, stop, step)]

        if idx < 0:
            idx += len(self)
        if idx < 0 or idx >= len(self):
            raise IndexError("Index out of range")

        return self._get_sync_point(idx)

    def __next__(self) -> SynchronizedPoint:
        """Get the next synchronized data point in the episode.

        Returns:
            SynchronizedPoint object containing synchronized data for the next timestep.

        Raises:
            StopIteration: When all timesteps have been processed.
        """
        if self._iter_idx >= len(self._episode_synced.observations):
            raise StopIteration
        sync_point = self._get_sync_point(self._iter_idx)
        self._iter_idx += 1
        return sync_point
